import torch
import torch.nn as nn
from .common import *


normalization = nn.BatchNorm2d


def conv(in_f, out_f, kernel_size, stride=1, bias=True, pad="zero"):
    if pad == "zero":
        return nn.Conv2d(
            in_f, out_f, kernel_size, stride, padding=(kernel_size - 1) / 2, bias=bias
        )
    elif pad == "reflection":
        layers = [
            nn.ReflectionPad2d((kernel_size - 1) / 2),
            nn.Conv2d(in_f, out_f, kernel_size, stride, padding=0, bias=bias),
        ]
        return nn.Sequential(*layers)


def get_texture_nets(
    inp=3,
    ratios=[32, 16, 8, 4, 2, 1],
    fill_noise=False,
    pad="zero",
    need_sigmoid=False,
    conv_num=8,
    upsample_mode="nearest",
):

    for i in range(len(ratios)):
        j = i + 1

        seq = nn.Sequential()

        tmp = nn.AvgPool2d(ratios[i], ratios[i])

        seq.add(tmp)
        if fill_noise:
            seq.add(GenNoise(inp))

        seq.add(conv(inp, conv_num, 3, pad=pad))
        seq.add(normalization(conv_num))
        seq.add(act())

        seq.add(conv(conv_num, conv_num, 3, pad=pad))
        seq.add(normalization(conv_num))
        seq.add(act())

        seq.add(conv(conv_num, conv_num, 1, pad=pad))
        seq.add(normalization(conv_num))
        seq.add(act())

        if i == 0:
            seq.add(nn.Upsample(scale_factor=2, mode=upsample_mode))
            cur = seq
        else:

            cur_temp = cur

            cur = nn.Sequential()

            # Batch norm before merging
            seq.add(normalization(conv_num))
            cur_temp.add(normalization(conv_num * (j - 1)))

            cur.add(Concat(1, cur_temp, seq))

            cur.add(conv(conv_num * j, conv_num * j, 3, pad=pad))
            cur.add(normalization(conv_num * j))
            cur.add(act())

            cur.add(conv(conv_num * j, conv_num * j, 3, pad=pad))
            cur.add(normalization(conv_num * j))
            cur.add(act())

            cur.add(conv(conv_num * j, conv_num * j, 1, pad=pad))
            cur.add(normalization(conv_num * j))
            cur.add(act())

            if i == len(ratios) - 1:
                cur.add(conv(conv_num * j, 3, 1, pad=pad))
            else:
                cur.add(nn.Upsample(scale_factor=2, mode=upsample_mode))

    model = cur
    if need_sigmoid:
        model.add(nn.Sigmoid())

    return model



INPUT_CHANNELS = 3
INPUT_HEIGHT = 48
INPUT_WIDTH = 48
INPUT_SIZE = (INPUT_WIDTH, INPUT_HEIGHT)
INPUT_SHAPE = (INPUT_CHANNELS, INPUT_HEIGHT, INPUT_WIDTH)
INPUT_SIZE_DICT = dict(
    height=INPUT_HEIGHT,
    width=INPUT_WIDTH,
)

N_AUGMENTATIONS = 9



import math
import torch
from torch import nn


def t_to_alpha_sigma(t):
    """Returns the scaling factors for the clean image and for the noise, given
    a timestep."""
    return torch.cos(t * math.pi / 2), torch.sin(t * math.pi / 2)


def alpha_sigma_to_log_snr(alpha, sigma):
    """Returns a log snr, given the scaling factors for the clean image and for
    the noise."""
    return torch.log(alpha**2 / sigma**2)


def log_snr_to_alpha_sigma(log_snr):
    """Returns the scaling factors for the clean image and for the noise, given
    the log SNR for a timestep."""
    return log_snr.sigmoid().sqrt(), log_snr.neg().sigmoid().sqrt()


def alpha_sigma_to_t(alpha, sigma):
    """Returns a timestep, given the scaling factors for the clean image and for
    the noise."""
    return torch.atan2(sigma, alpha) / math.pi * 2


def get_ddpm_schedule(ddpm_t):
    """Returns timesteps for the noise schedule from the DDPM paper."""
    log_snr = -torch.special.expm1(1e-4 + 10 * ddpm_t**2).log()
    alpha, sigma = log_snr_to_alpha_sigma(log_snr)
    return alpha_sigma_to_t(alpha, sigma)


class ResidualBlock(nn.Module):
    def __init__(self, main, skip=None):
        super().__init__()
        self.skip = skip if skip else nn.Identity()
        self.main = nn.Sequential(*main)

    def forward(self, input):
        return self.main(input) + self.skip(input)


class ResConvBlock(ResidualBlock):
    def __init__(self, c_in, c_mid, c_out, is_last=False):
        skip = None if c_in == c_out else nn.Conv2d(c_in, c_out, 1, bias=False)
        super().__init__(
            [
                nn.Conv2d(c_in, c_mid, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(c_mid, c_out, 3, padding=1),
                nn.ReLU(inplace=True) if not is_last else nn.Identity(),
            ],
            skip,
        )


class SelfAttention2d(nn.Module):
    def __init__(self, c_in, n_head=1):
        super().__init__()
        assert c_in % n_head == 0
        self.n_head = n_head
        self.qkv_proj = nn.Conv2d(c_in, c_in * 3, 1)
        self.out_proj = nn.Conv2d(c_in, c_in, 1)

    def forward(self, input):
        n, c, h, w = input.shape
        qkv = self.qkv_proj(input)
        qkv = qkv.view([n, self.n_head * 3, c // self.n_head, h * w]).transpose(2, 3)
        q, k, v = qkv.chunk(3, dim=1)
        scale = k.shape[3] ** -0.25
        att = ((q * scale) @ (k.transpose(2, 3) * scale)).softmax(3)
        y = (att @ v).transpose(2, 3).contiguous().view([n, c, h, w])
        return input + self.out_proj(y)


class SkipBlock(nn.Module):
    def __init__(self, main, skip=None):
        super().__init__()
        self.main = nn.Sequential(*main)
        self.skip = skip if skip else nn.Identity()

    def forward(self, input):
        return torch.cat([self.skip(input), self.main(input)], dim=1)


class FourierFeatures(nn.Module):
    def __init__(self, in_features, out_features, std=1.0):
        super().__init__()
        assert out_features % 2 == 0
        self.weight = nn.Parameter(torch.randn([out_features // 2, in_features]) * std)

    def forward(self, input):
        f = 2 * math.pi * input @ self.weight.T
        return torch.cat([f.cos(), f.sin()], dim=-1)


def expand_to_planes(input, shape):
    return input[..., None, None].repeat([1, 1, shape[2], shape[3]])


class WikiArt256Model(nn.Module):
    def __init__(self):
        super().__init__()
        self.shape = (3, 256, 256)
        self.min_t = get_ddpm_schedule(torch.tensor(0.0)).item()
        self.max_t = get_ddpm_schedule(torch.tensor(1.0)).item()

        c = 128  # The base channel count
        cs = [c // 2, c, c * 2, c * 2, c * 4, c * 4, c * 8]

        self.timestep_embed = FourierFeatures(1, 16, std=0.2)
        self.down = nn.AvgPool2d(2)
        self.up = nn.Upsample(scale_factor=2, mode="nearest")

        self.net = nn.Sequential(  # 256x256
            ResConvBlock(3 + 16, cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], cs[0]),
            SkipBlock(
                [
                    self.down,  # 256x256 -> 128x128
                    ResConvBlock(cs[0], cs[1], cs[1]),
                    ResConvBlock(cs[1], cs[1], cs[1]),
                    ResConvBlock(cs[1], cs[1], cs[1]),
                    ResConvBlock(cs[1], cs[1], cs[1]),
                    SkipBlock(
                        [
                            self.down,  # 128x128 -> 64x64
                            ResConvBlock(cs[1], cs[2], cs[2]),
                            ResConvBlock(cs[2], cs[2], cs[2]),
                            ResConvBlock(cs[2], cs[2], cs[2]),
                            ResConvBlock(cs[2], cs[2], cs[2]),
                            SkipBlock(
                                [
                                    self.down,  # 64x64 -> 32x32
                                    ResConvBlock(cs[2], cs[3], cs[3]),
                                    ResConvBlock(cs[3], cs[3], cs[3]),
                                    ResConvBlock(cs[3], cs[3], cs[3]),
                                    ResConvBlock(cs[3], cs[3], cs[3]),
                                    SkipBlock(
                                        [
                                            self.down,  # 32x32 -> 16x16
                                            ResConvBlock(cs[3], cs[4], cs[4]),
                                            SelfAttention2d(cs[4], cs[4] // 128),
                                            ResConvBlock(cs[4], cs[4], cs[4]),
                                            SelfAttention2d(cs[4], cs[4] // 128),
                                            ResConvBlock(cs[4], cs[4], cs[4]),
                                            SelfAttention2d(cs[4], cs[4] // 128),
                                            ResConvBlock(cs[4], cs[4], cs[4]),
                                            SelfAttention2d(cs[4], cs[4] // 128),
                                            SkipBlock(
                                                [
                                                    self.down,  # 16x16 -> 8x8
                                                    ResConvBlock(cs[4], cs[5], cs[5]),
                                                    SelfAttention2d(
                                                        cs[5], cs[5] // 128
                                                    ),
                                                    ResConvBlock(cs[5], cs[5], cs[5]),
                                                    SelfAttention2d(
                                                        cs[5], cs[5] // 128
                                                    ),
                                                    ResConvBlock(cs[5], cs[5], cs[5]),
                                                    SelfAttention2d(
                                                        cs[5], cs[5] // 128
                                                    ),
                                                    ResConvBlock(cs[5], cs[5], cs[5]),
                                                    SelfAttention2d(
                                                        cs[5], cs[5] // 128
                                                    ),
                                                    SkipBlock(
                                                        [
                                                            self.down,  # 8x8 -> 4x4
                                                            ResConvBlock(
                                                                cs[5], cs[6], cs[6]
                                                            ),
                                                            SelfAttention2d(
                                                                cs[6], cs[6] // 128
                                                            ),
                                                            ResConvBlock(
                                                                cs[6], cs[6], cs[6]
                                                            ),
                                                            SelfAttention2d(
                                                                cs[6], cs[6] // 128
                                                            ),
                                                            ResConvBlock(
                                                                cs[6], cs[6], cs[6]
                                                            ),
                                                            SelfAttention2d(
                                                                cs[6], cs[6] // 128
                                                            ),
                                                            ResConvBlock(
                                                                cs[6], cs[6], cs[6]
                                                            ),
                                                            SelfAttention2d(
                                                                cs[6], cs[6] // 128
                                                            ),
                                                            ResConvBlock(
                                                                cs[6], cs[6], cs[6]
                                                            ),
                                                            SelfAttention2d(
                                                                cs[6], cs[6] // 128
                                                            ),
                                                            ResConvBlock(
                                                                cs[6], cs[6], cs[6]
                                                            ),
                                                            SelfAttention2d(
                                                                cs[6], cs[6] // 128
                                                            ),
                                                            ResConvBlock(
                                                                cs[6], cs[6], cs[6]
                                                            ),
                                                            SelfAttention2d(
                                                                cs[6], cs[6] // 128
                                                            ),
                                                            ResConvBlock(
                                                                cs[6], cs[6], cs[5]
                                                            ),
                                                            SelfAttention2d(
                                                                cs[5], cs[5] // 128
                                                            ),
                                                            self.up,  # 4x4 -> 8x8
                                                        ]
                                                    ),
                                                    ResConvBlock(
                                                        cs[5] * 2, cs[5], cs[5]
                                                    ),
                                                    SelfAttention2d(
                                                        cs[5], cs[5] // 128
                                                    ),
                                                    ResConvBlock(cs[5], cs[5], cs[5]),
                                                    SelfAttention2d(
                                                        cs[5], cs[5] // 128
                                                    ),
                                                    ResConvBlock(cs[5], cs[5], cs[5]),
                                                    SelfAttention2d(
                                                        cs[5], cs[5] // 128
                                                    ),
                                                    ResConvBlock(cs[5], cs[5], cs[4]),
                                                    SelfAttention2d(
                                                        cs[4], cs[4] // 128
                                                    ),
                                                    self.up,  # 8x8 -> 16x16
                                                ]
                                            ),
                                            ResConvBlock(cs[4] * 2, cs[4], cs[4]),
                                            SelfAttention2d(cs[4], cs[4] // 128),
                                            ResConvBlock(cs[4], cs[4], cs[4]),
                                            SelfAttention2d(cs[4], cs[4] // 128),
                                            ResConvBlock(cs[4], cs[4], cs[4]),
                                            SelfAttention2d(cs[4], cs[4] // 128),
                                            ResConvBlock(cs[4], cs[4], cs[3]),
                                            SelfAttention2d(cs[3], cs[3] // 128),
                                            self.up,  # 16x16 -> 32x32
                                        ]
                                    ),
                                    ResConvBlock(cs[3] * 2, cs[3], cs[3]),
                                    ResConvBlock(cs[3], cs[3], cs[3]),
                                    ResConvBlock(cs[3], cs[3], cs[3]),
                                    ResConvBlock(cs[3], cs[3], cs[2]),
                                    self.up,  # 32x32 -> 64x64
                                ]
                            ),
                            ResConvBlock(cs[2] * 2, cs[2], cs[2]),
                            ResConvBlock(cs[2], cs[2], cs[2]),
                            ResConvBlock(cs[2], cs[2], cs[2]),
                            ResConvBlock(cs[2], cs[2], cs[1]),
                            self.up,  # 64x64 -> 128x128
                        ]
                    ),
                    ResConvBlock(cs[1] * 2, cs[1], cs[1]),
                    ResConvBlock(cs[1], cs[1], cs[1]),
                    ResConvBlock(cs[1], cs[1], cs[1]),
                    ResConvBlock(cs[1], cs[1], cs[0]),
                    self.up,  # 128x128 -> 256x256
                ]
            ),
            ResConvBlock(cs[0] * 2, cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], 3, is_last=True),
        )

    def forward(self, input, t):
        log_snr = alpha_sigma_to_log_snr(*t_to_alpha_sigma(t))
        timestep_embed = expand_to_planes(
            self.timestep_embed(log_snr[:, None]), input.shape
        )
        return self.net(torch.cat([input, timestep_embed], dim=1))



import torch
from torch.nn import functional as F

from .interface import LossInterface
from perceptor import models


class SimulacraAesthetic(LossInterface):
    def __init__(self, model_name="ViT-L-14", aesthetic_target=10):
        """
        Simulacra aesthetic loss based on clip linear regression probe that predicts the aesthetic rating of an image.

        Args:
            model_name (str): Name of CLIP model. Available models are:
                - ViT-B-32
                - ViT-B-16
                - ViT-L-14
                - RN50
                - RN101
                - RN50x4
                - RN50x16
                - RN50x64
                - ViT-L-14-336
            aesthetic_target (int): Target asthetic rating of the image (1-10).
        """
        super().__init__()
        self.aesthetic_target = torch.nn.Parameter(
            torch.as_tensor(aesthetic_target).float(), requires_grad=False
        )
        self.model = models.SimulacraAesthetic(model_name)
        if model_name in ("ViT-L-14", "ViT-L-14-336"):
            self.multiplier = 0.00001
        else:
            self.multiplier = 0.001

    def forward(self, images):
        predicted_aesthetic_rating = self.model(images)
        return self.multiplier * F.mse_loss(
            predicted_aesthetic_rating,
            self.aesthetic_target.view(-1, 1),
        )


def test_simulacra_aesthetic_loss():
    loss = SimulacraAesthetic("ViT-B-32").cuda()
    loss(torch.randn(1, 3, 256, 256).cuda())



import torch
import torch.nn.functional as F

from perceptor.transforms.interface import TransformInterface
from perceptor import models
from perceptor.transforms.resize import resize


class SuperResolution(TransformInterface):
    def __init__(self, name="x4", half=False):
        super().__init__()
        self.name = name
        self.model = models.SuperResolution(name, half)

    def encode(self, images):
        return self.model.upsample(images)

    def decode(self, upsampled_images, size=None):
        if size is None:
            size = (
                torch.tensor(upsampled_images.shape[-2:]) // self.model.scale
            ).tolist()
        return resize(
            upsampled_images,
            out_shape=size,
        )



import torch

from perceptor import models
from perceptor.losses.interface import LossInterface


class OpenCLIP(LossInterface):
    def __init__(
        self,
        architecture="ViT-H-14",
        weights="laion2b_s32b_b79k",
    ):
        """
        Args:
            architecture (str): name of the clip model
            weights (str): name of the weights

            Available weight/model combinations are (in order of relevance):
            - ("ViT-H-14", "laion2b_s32b_b79k") (78.0%)
            - ("ViT-g-14", "laion2b_s12b_b42k") (76.6%)
            - ("ViT-L-14", "laion2b_s32b_b82k") (75.3%)
            - ("ViT-B-32", "laion2b_s34b_b79k") (66.6%)
            - ("ViT-B-16-plus-240", "laion400m_e32") (69.2%)
            - ("ViT-B-32", "laion2b_e16") (65.7%)
            - ("ViT-B-16", "laion400m_e32") (67.0%)
            - ("ViT-B-32", "laion400m_e32") (62.9%)
            - ("ViT-L-14", "laion400m_e32") (72.8%)
            - ("RN101", "yfcc15m") (34.8%)
            - ("RN50", "yfcc15m") (32.7%)
            - ("RN50", "cc12m") (36.45%)
            - ("RN50-quickgelu", "openai") (59.6%)
            - ("RN101-quickgelu", "openai")
            - ("RN50x4", "openai")
            - ("RN50x16", "openai")
            - ("RN50x64", "openai")
            - ("ViT-B-32-quickgelu", "openai") (63.3%)
            - ("ViT-B-16", "openai") (68.3%)
            - ("ViT-L-14", "openai") (75.6%)
            - ("ViT-L-14-336", "openai") (76.6%)
        """
        super().__init__()
        self.architecture = architecture
        self.weights = weights
        self.model = models.OpenCLIP(architecture, weights)
        self.encodings = None
        self.weights = None

    @property
    def device(self):
        return next(iter(self.model.parameters())).device

    def add_texts_(self, texts, weights=None):
        return self.add_encodings_(self.model.encode_texts(texts), weights)

    def add_images_(self, images, weights=None):
        return self.add_encodings_(self.model.encode_images(images), weights)

    def add_encodings_(
        self,
        encodings,
        weights=None,
    ):
        if isinstance(weights, list) or isinstance(weights, tuple):
            weights = torch.tensor(weights)
        elif weights is None:
            weights = torch.ones_like(encodings[:, 0])

        if self.encodings is None:
            self.encodings = torch.nn.Parameter(
                encodings.to(self.device), requires_grad=False
            )
            self.weights = torch.nn.Parameter(
                weights.to(self.device),
                requires_grad=False,
            )
        else:
            self.encodings = torch.nn.Parameter(
                torch.cat([self.encodings, encodings.to(self.device)]),
                requires_grad=False,
            )
            self.weights = torch.nn.Parameter(
                torch.cat([self.weights, weights.to(self.device)]),
                requires_grad=False,
            )
        return self

    def forward(self, images):
        image_encodings = self.model.encode_images(images)
        spherical_distance = (
            (image_encodings[:, None] - self.encodings[None, :])
            .norm(dim=2)
            .div(2)
            .arcsin()
            .square()
            .mul(2)
        )
        return (spherical_distance * self.weights).mean()


def test_open_clip():
    loss = (
        OpenCLIP()
        .add_texts_(["hello", "world"])
        .add_images_(torch.randn(1, 3, 256, 256))
    )
    loss(torch.randn(1, 3, 256, 256))



import torch
import torch.nn.functional as F

from perceptor import utils
from .load import load


@utils.cache
class RuCLIP(torch.nn.Module):
    def __init__(self, name):
        super().__init__()
        self.name = name
        start_device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )
        self.model, self.clip_processor = load(name, start_device)
        self.model.requires_grad_(False).eval()

        # softmax on cpu does not support half precision
        if not torch.cuda.is_available():
            self.model.float()

    @property
    def device(self):
        return next(iter(self.parameters())).device

    def encode_texts(self, text_prompts):
        inputs = self.clip_processor(
            text=text_prompts, return_tensors="pt", padding=True
        )
        input_ids = inputs["input_ids"].to(self.device)
        return F.normalize(self.model.encode_text(input_ids))

    def encode_images(self, images):
        return F.normalize(
            self.model.encode_image(
                self.clip_processor.image_transform(images.to(self.device))
            )
        )

    def forward(self, _):
        raise NotImplementedError



import torch
from basicsr.utils.download_util import load_file_from_url

from perceptor.losses.interface import LossInterface
from .unet_discriminator_sn import UNetDiscriminatorSN


checkpoints = {
    "RealESRGAN_x4plus_netD": "https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.2.3/RealESRGAN_x4plus_netD.pth",
}


class SuperResolutionDiscriminator(LossInterface):
    def __init__(self, name="RealESRGAN_x4plus_netD"):
        super().__init__()
        self.name = name

        checkpoint_path = load_file_from_url(
            checkpoints[self.name],
            "models",
        )

        self.model = UNetDiscriminatorSN(num_in_ch=3, num_feat=64, skip_connection=True)
        self.model.requires_grad_(False)
        weights = torch.load(checkpoint_path, map_location=torch.device("cpu"))
        self.model.load_state_dict(weights["params"], strict=True)

    def forward(self, images):
        return -self.model(images).mean() * 0.001



import torch
import torch.nn as nn
from functools import partial

from perceptor.models.latent_diffusion.ldm.modules.x_transformer import (
    Encoder,
    TransformerWrapper,
)  # TODO: can we directly rely on lucidrains code and simply add this as a reuirement? --> test


class AbstractEncoder(nn.Module):
    def __init__(self):
        super().__init__()

    def encode(self, *args, **kwargs):
        raise NotImplementedError


class ClassEmbedder(nn.Module):
    def __init__(self, embed_dim, n_classes=1000, key="class"):
        super().__init__()
        self.key = key
        self.embedding = nn.Embedding(n_classes, embed_dim)

    def forward(self, batch, key=None):
        if key is None:
            key = self.key
        # this is for use in crossattn
        c = batch[key][:, None]
        c = self.embedding(c)
        return c


class TransformerEmbedder(AbstractEncoder):
    """Some transformer encoder layers"""

    def __init__(self, n_embed, n_layer, vocab_size, max_seq_len=77, device="cuda"):
        super().__init__()
        self.device = device
        self.transformer = TransformerWrapper(
            num_tokens=vocab_size,
            max_seq_len=max_seq_len,
            attn_layers=Encoder(dim=n_embed, depth=n_layer),
        )

    def forward(self, tokens):
        tokens = tokens.to(self.device)  # meh
        z = self.transformer(tokens, return_embeddings=True)
        return z

    def encode(self, x):
        return self(x)


class BERTTokenizer(AbstractEncoder):
    """Uses a pretrained BERT tokenizer by huggingface. Vocab size: 30522 (?)"""

    def __init__(self, device="cuda", vq_interface=True, max_length=77):
        super().__init__()
        from transformers import BertTokenizerFast  # TODO: add to reuquirements

        self.tokenizer = BertTokenizerFast.from_pretrained("bert-base-uncased")
        self.device = device
        self.vq_interface = vq_interface
        self.max_length = max_length

    def forward(self, text):
        batch_encoding = self.tokenizer(
            text,
            truncation=True,
            max_length=self.max_length,
            return_length=True,
            return_overflowing_tokens=False,
            padding="max_length",
            return_tensors="pt",
        )
        tokens = batch_encoding["input_ids"].to(self.device)
        return tokens

    @torch.no_grad()
    def encode(self, text):
        tokens = self(text)
        if not self.vq_interface:
            return tokens
        return None, None, [None, None, tokens]

    def decode(self, text):
        return text


class BERTEmbedder(AbstractEncoder):
    """Uses the BERT tokenizr model and add some transformer encoder layers"""

    def __init__(
        self,
        n_embed,
        n_layer,
        vocab_size=30522,
        max_seq_len=77,
        device="cuda",
        use_tokenizer=True,
        embedding_dropout=0.0,
    ):
        super().__init__()
        self.use_tknz_fn = use_tokenizer
        if self.use_tknz_fn:
            self.tknz_fn = BERTTokenizer(vq_interface=False, max_length=max_seq_len)
        self.device = device
        self.transformer = TransformerWrapper(
            num_tokens=vocab_size,
            max_seq_len=max_seq_len,
            attn_layers=Encoder(dim=n_embed, depth=n_layer),
            emb_dropout=embedding_dropout,
        )

    def forward(self, text):
        if self.use_tknz_fn:
            tokens = self.tknz_fn(text)  # .to(self.device)
        else:
            tokens = text
        z = self.transformer(tokens, return_embeddings=True)
        return z

    def encode(self, text):
        # output of length 77
        return self(text)


class SpatialRescaler(nn.Module):
    def __init__(
        self,
        n_stages=1,
        method="bilinear",
        multiplier=0.5,
        in_channels=3,
        out_channels=None,
        bias=False,
    ):
        super().__init__()
        self.n_stages = n_stages
        assert self.n_stages >= 0
        assert method in [
            "nearest",
            "linear",
            "bilinear",
            "trilinear",
            "bicubic",
            "area",
        ]
        self.multiplier = multiplier
        self.interpolator = partial(torch.nn.functional.interpolate, mode=method)
        self.remap_output = out_channels is not None
        if self.remap_output:
            print(
                f"Spatial Rescaler mapping from {in_channels} to {out_channels} channels after resizing."
            )
            self.channel_mapper = nn.Conv2d(in_channels, out_channels, 1, bias=bias)

    def forward(self, x):
        for stage in range(self.n_stages):
            x = self.interpolator(x, scale_factor=self.multiplier)

        if self.remap_output:
            x = self.channel_mapper(x)
        return x

    def encode(self, x):
        return self(x)



from hashlib import sha256
from pathlib import Path

from torch import hub

from .config import load_config


CONFIG_DIR = Path(__file__).resolve().parent / "pretrained_configs"


def list_configs():
    return sorted(path.stem for path in CONFIG_DIR.glob("*.json"))


def get_config(name):
    config_path = CONFIG_DIR / (name + ".json")
    if config_path.is_file():
        return load_config(config_path)


def download_checkpoint(config):
    dest_dir = Path(hub.get_dir()) / "cloob"
    dest_dir.mkdir(parents=True, exist_ok=True)
    url = config["url"]
    filename = url.rpartition("/")[2]
    model_hash = filename.rpartition("-")[2].partition(".")[0]
    dest_file = dest_dir / filename
    if not dest_file.exists():
        hub.download_url_to_file(url, str(dest_file), hash_prefix=model_hash)
    if sha256(open(dest_file, "rb").read()).hexdigest() == model_hash:
        return str(dest_file)
    raise RuntimeError(
        f"Model has been downloaded to {dest_file!s} but the SHA256 checksum does not not match"
    )



import math
from abc import abstractmethod

import torch as th
import torch.nn as nn
import torch.nn.functional as F

from .fp16_util import convert_module_to_f16, convert_module_to_f32
from .nn import (
    avg_pool_nd,
    conv_nd,
    linear,
    normalization,
    timestep_embedding,
    zero_module,
    checkpoint,
)

from inspect import isfunction
from einops import rearrange, repeat
from torch import nn, einsum

# dummy replace
# def convert_module_to_f16(x):
#    pass

# def convert_module_to_f32(x):
#    pass


def exists(val):
    return val is not None


def default(val, d):
    if exists(val):
        return val
    return d() if isfunction(d) else d


class GroupNorm32(nn.GroupNorm):
    def forward(self, x):
        return super().forward(x.float()).type(x.dtype)


def Normalize(in_channels):
    return GroupNorm32(num_groups=32, num_channels=in_channels, eps=1e-6, affine=True)


class LayerNorm(nn.LayerNorm):
    """
    Implementation that supports fp16 inputs but fp32 gains/biases.
    """

    def forward(self, x: th.Tensor):
        return super().forward(x.float()).to(x.dtype)


class Linear(nn.Linear):
    """
    Implementation that supports fp16 inputs but fp32 gains/biases.
    """

    def forward(self, x: th.Tensor):
        return super().forward(x.float()).to(x.dtype)


class GEGLU(nn.Module):
    def __init__(self, dim_in, dim_out):
        super().__init__()
        self.proj = Linear(dim_in, dim_out * 2)

    def forward(self, x):
        x, gate = self.proj(x).chunk(2, dim=-1)
        return x * F.gelu(gate)


class FeedForward(nn.Module):
    def __init__(self, dim, dim_out=None, mult=4, glu=False, dropout=0.0):
        super().__init__()
        inner_dim = int(dim * mult)
        dim_out = default(dim_out, dim)
        project_in = (
            nn.Sequential(Linear(dim, inner_dim), nn.GELU())
            if not glu
            else GEGLU(dim, inner_dim)
        )

        self.net = nn.Sequential(
            project_in, nn.Dropout(dropout), Linear(inner_dim, dim_out)
        )

    def forward(self, x):
        return self.net(x)


class CrossAttention(nn.Module):
    def __init__(self, query_dim, context_dim=None, heads=8, dim_head=64, dropout=0.0):
        super().__init__()
        inner_dim = dim_head * heads
        context_dim = default(context_dim, query_dim)

        self.scale = dim_head**-0.5
        self.heads = heads

        self.to_q = Linear(query_dim, inner_dim, bias=False)
        self.to_k = Linear(context_dim, inner_dim, bias=False)
        self.to_v = Linear(context_dim, inner_dim, bias=False)

        self.to_out = nn.Sequential(Linear(inner_dim, query_dim), nn.Dropout(dropout))

    def forward(self, x, context=None, mask=None):
        h = self.heads

        q = self.to_q(x)
        context = default(context, x)

        k = self.to_k(context)
        v = self.to_v(context)

        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> (b h) n d", h=h), (q, k, v))

        sim = einsum("b i d, b j d -> b i j", q, k) * self.scale

        if exists(mask):
            mask = rearrange(mask, "b ... -> b (...)")
            max_neg_value = -torch.finfo(sim.dtype).max
            mask = repeat(mask, "b j -> (b h) () j", h=h)
            sim.masked_fill_(~mask, max_neg_value)

        # attention, what we cannot get enough of
        attn = sim.softmax(dim=-1)

        out = einsum("b i j, b j d -> b i d", attn, v)
        out = rearrange(out, "(b h) n d -> b n (h d)", h=h)
        return self.to_out(out)


class BasicTransformerBlock(nn.Module):
    def __init__(
        self,
        dim,
        n_heads,
        d_head,
        dropout=0.0,
        context_dim=None,
        gated_ff=True,
        checkpoint=True,
    ):
        super().__init__()
        self.attn1 = CrossAttention(
            query_dim=dim, heads=n_heads, dim_head=d_head, dropout=dropout
        )  # is a self-attention
        self.ff = FeedForward(dim, dropout=dropout, glu=gated_ff)
        self.attn2 = CrossAttention(
            query_dim=dim,
            context_dim=context_dim,
            heads=n_heads,
            dim_head=d_head,
            dropout=dropout,
        )  # is self-attn if context is none
        self.norm1 = LayerNorm(dim)
        self.norm2 = LayerNorm(dim)
        self.norm3 = LayerNorm(dim)
        self.checkpoint = checkpoint

    def forward(self, x, context=None):
        return checkpoint(
            self._forward, (x, context), self.parameters(), self.checkpoint
        )

    def _forward(self, x, context=None):
        x = self.attn1(self.norm1(x)) + x
        x = self.attn2(self.norm2(x), context=context) + x
        x = self.ff(self.norm3(x)) + x
        return x


class SpatialTransformer(nn.Module):
    """
    Transformer block for image-like data.
    First, project the input (aka embedding)
    and reshape to b, t, d.
    Then apply standard transformer action.
    Finally, reshape to image
    """

    def __init__(
        self, in_channels, n_heads, d_head, depth=1, dropout=0.0, context_dim=None
    ):
        super().__init__()
        self.in_channels = in_channels
        inner_dim = n_heads * d_head
        self.norm = Normalize(in_channels)

        self.proj_in = nn.Conv2d(
            in_channels, inner_dim, kernel_size=1, stride=1, padding=0
        )

        self.transformer_blocks = nn.ModuleList(
            [
                BasicTransformerBlock(
                    inner_dim, n_heads, d_head, dropout=dropout, context_dim=context_dim
                )
                for d in range(depth)
            ]
        )

        self.proj_out = zero_module(
            nn.Conv2d(inner_dim, in_channels, kernel_size=1, stride=1, padding=0)
        )

    def forward(self, x, context=None):
        # note: if no context is given, cross-attention defaults to self-attention
        b, c, h, w = x.shape
        x_in = x
        x = self.norm(x)
        x = self.proj_in(x)
        x = rearrange(x, "b c h w -> b (h w) c")
        for block in self.transformer_blocks:
            x = block(x, context=context)
        x = rearrange(x, "b (h w) c -> b c h w", h=h, w=w)
        x = self.proj_out(x)
        return x + x_in


class TimestepBlock(nn.Module):
    """
    Any module where forward() takes timestep embeddings as a second argument.
    """

    @abstractmethod
    def forward(self, x, emb):
        """
        Apply the module to `x` given `emb` timestep embeddings.
        """


class TimestepEmbedSequential(nn.Sequential, TimestepBlock):
    """
    A sequential module that passes timestep embeddings to the children that
    support it as an extra input.
    """

    def forward(self, x, emb, context=None):
        for layer in self:
            if isinstance(layer, TimestepBlock):
                x = layer(x, emb)
            elif isinstance(layer, SpatialTransformer):
                x = layer(x, context)
            else:
                x = layer(x)
        return x


class Upsample(nn.Module):
    """
    An upsampling layer with an optional convolution.
    :param channels: channels in the inputs and outputs.
    :param use_conv: a bool determining if a convolution is applied.
    :param dims: determines if the signal is 1D, 2D, or 3D. If 3D, then
                 upsampling occurs in the inner-two dimensions.
    """

    def __init__(self, channels, use_conv, dims=2, out_channels=None, padding=1):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.dims = dims
        if use_conv:
            self.conv = conv_nd(
                dims, self.channels, self.out_channels, 3, padding=padding
            )

    def forward(self, x):
        assert x.shape[1] == self.channels
        if self.dims == 3:
            x = F.interpolate(
                x, (x.shape[2], x.shape[3] * 2, x.shape[4] * 2), mode="nearest"
            )
        else:
            x = F.interpolate(x, scale_factor=2, mode="nearest")
        if self.use_conv:
            x = self.conv(x)
        return x


class Downsample(nn.Module):
    """
    A downsampling layer with an optional convolution.
    :param channels: channels in the inputs and outputs.
    :param use_conv: a bool determining if a convolution is applied.
    :param dims: determines if the signal is 1D, 2D, or 3D. If 3D, then
                 downsampling occurs in the inner-two dimensions.
    """

    def __init__(self, channels, use_conv, dims=2, out_channels=None, padding=1):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.dims = dims
        stride = 2 if dims != 3 else (1, 2, 2)
        if use_conv:
            self.op = conv_nd(
                dims,
                self.channels,
                self.out_channels,
                3,
                stride=stride,
                padding=padding,
            )
        else:
            assert self.channels == self.out_channels
            self.op = avg_pool_nd(dims, kernel_size=stride, stride=stride)

    def forward(self, x):
        assert x.shape[1] == self.channels
        return self.op(x)


class ResBlock(TimestepBlock):
    """
    A residual block that can optionally change the number of channels.
    :param channels: the number of input channels.
    :param emb_channels: the number of timestep embedding channels.
    :param dropout: the rate of dropout.
    :param out_channels: if specified, the number of out channels.
    :param use_conv: if True and out_channels is specified, use a spatial
        convolution instead of a smaller 1x1 convolution to change the
        channels in the skip connection.
    :param dims: determines if the signal is 1D, 2D, or 3D.
    :param use_checkpoint: if True, use gradient checkpointing on this module.
    :param up: if True, use this block for upsampling.
    :param down: if True, use this block for downsampling.
    """

    def __init__(
        self,
        channels,
        emb_channels,
        dropout,
        out_channels=None,
        use_conv=False,
        use_scale_shift_norm=False,
        dims=2,
        use_checkpoint=False,
        up=False,
        down=False,
    ):
        super().__init__()
        self.channels = channels
        self.emb_channels = emb_channels
        self.dropout = dropout
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.use_checkpoint = use_checkpoint
        self.use_scale_shift_norm = use_scale_shift_norm

        self.in_layers = nn.Sequential(
            normalization(channels),
            nn.SiLU(),
            conv_nd(dims, channels, self.out_channels, 3, padding=1),
        )

        self.updown = up or down

        if up:
            self.h_upd = Upsample(channels, False, dims)
            self.x_upd = Upsample(channels, False, dims)
        elif down:
            self.h_upd = Downsample(channels, False, dims)
            self.x_upd = Downsample(channels, False, dims)
        else:
            self.h_upd = self.x_upd = nn.Identity()

        self.emb_layers = nn.Sequential(
            nn.SiLU(),
            linear(
                emb_channels,
                2 * self.out_channels if use_scale_shift_norm else self.out_channels,
            ),
        )
        self.out_layers = nn.Sequential(
            normalization(self.out_channels),
            nn.SiLU(),
            nn.Dropout(p=dropout),
            zero_module(
                conv_nd(dims, self.out_channels, self.out_channels, 3, padding=1)
            ),
        )

        if self.out_channels == channels:
            self.skip_connection = nn.Identity()
        elif use_conv:
            self.skip_connection = conv_nd(
                dims, channels, self.out_channels, 3, padding=1
            )
        else:
            self.skip_connection = conv_nd(dims, channels, self.out_channels, 1)

    def forward(self, x, emb):
        """
        Apply the block to a Tensor, conditioned on a timestep embedding.
        :param x: an [N x C x ...] Tensor of features.
        :param emb: an [N x emb_channels] Tensor of timestep embeddings.
        :return: an [N x C x ...] Tensor of outputs.
        """
        return checkpoint(
            self._forward, (x, emb), self.parameters(), self.use_checkpoint
        )

    def _forward(self, x, emb):
        if self.updown:
            in_rest, in_conv = self.in_layers[:-1], self.in_layers[-1]
            h = in_rest(x)
            h = self.h_upd(h)
            x = self.x_upd(x)
            h = in_conv(h)
        else:
            h = self.in_layers(x)
        emb_out = self.emb_layers(emb).type(h.dtype)
        while len(emb_out.shape) < len(h.shape):
            emb_out = emb_out[..., None]
        if self.use_scale_shift_norm:
            out_norm, out_rest = self.out_layers[0], self.out_layers[1:]
            scale, shift = th.chunk(emb_out, 2, dim=1)
            h = out_norm(h) * (1 + scale) + shift
            h = out_rest(h)
        else:
            h = h + emb_out
            h = self.out_layers(h)
        return self.skip_connection(x) + h


class AttentionBlock(nn.Module):
    """
    An attention block that allows spatial positions to attend to each other.
    Originally ported from here, but adapted to the N-d case.
    https://github.com/hojonathanho/diffusion/blob/1e0dceb3b3495bbe19116a5e1b3596cd0706c543/diffusion_tf/models/unet.py#L66.
    """

    def __init__(
        self,
        channels,
        num_heads=1,
        num_head_channels=-1,
        use_checkpoint=False,
        use_new_attention_order=False,
    ):
        super().__init__()
        self.channels = channels
        if num_head_channels == -1:
            self.num_heads = num_heads
        else:
            assert (
                channels % num_head_channels == 0
            ), f"q,k,v channels {channels} is not divisible by num_head_channels {num_head_channels}"
            self.num_heads = channels // num_head_channels
        self.use_checkpoint = use_checkpoint
        self.norm = normalization(channels)
        self.qkv = conv_nd(1, channels, channels * 3, 1)
        if use_new_attention_order:
            # split qkv before split heads
            self.attention = QKVAttention(self.num_heads)
        else:
            # split heads before split qkv
            self.attention = QKVAttentionLegacy(self.num_heads)

        self.proj_out = zero_module(conv_nd(1, channels, channels, 1))

    def forward(self, x):
        return checkpoint(
            self._forward, (x,), self.parameters(), True
        )  # TODO: check checkpoint usage, is True # TODO: fix the .half call!!!
        # return pt_checkpoint(self._forward, x)  # pytorch

    def _forward(self, x):
        b, c, *spatial = x.shape
        x = x.reshape(b, c, -1)
        qkv = self.qkv(self.norm(x))
        h = self.attention(qkv)
        h = self.proj_out(h)
        return (x + h).reshape(b, c, *spatial)


def count_flops_attn(model, _x, y):
    """
    A counter for the `thop` package to count the operations in an
    attention operation.
    Meant to be used like:
        macs, params = thop.profile(
            model,
            inputs=(inputs, timestamps),
            custom_ops={QKVAttention: QKVAttention.count_flops},
        )
    """
    b, c, *spatial = y[0].shape
    num_spatial = int(np.prod(spatial))
    # We perform two matmuls with the same number of ops.
    # The first computes the weight matrix, the second computes
    # the combination of the value vectors.
    matmul_ops = 2 * b * (num_spatial**2) * c
    model.total_ops += th.DoubleTensor([matmul_ops])


class QKVAttention(nn.Module):
    """
    A module which performs QKV attention and splits in a different order.
    """

    def __init__(self, n_heads):
        super().__init__()
        self.n_heads = n_heads

    def forward(self, qkv):
        """
        Apply QKV attention.
        :param qkv: an [N x (3 * H * C) x T] tensor of Qs, Ks, and Vs.
        :return: an [N x (H * C) x T] tensor after attention.
        """
        bs, width, length = qkv.shape
        assert width % (3 * self.n_heads) == 0
        ch = width // (3 * self.n_heads)
        q, k, v = qkv.chunk(3, dim=1)
        scale = 1 / math.sqrt(math.sqrt(ch))
        weight = th.einsum(
            "bct,bcs->bts",
            (q * scale).view(bs * self.n_heads, ch, length),
            (k * scale).view(bs * self.n_heads, ch, length),
        )  # More stable with f16 than dividing afterwards
        weight = th.softmax(weight.float(), dim=-1).type(weight.dtype)
        a = th.einsum("bts,bcs->bct", weight, v.reshape(bs * self.n_heads, ch, length))
        return a.reshape(bs, -1, length)

    @staticmethod
    def count_flops(model, _x, y):
        return count_flops_attn(model, _x, y)


class UNetModel(nn.Module):
    """
    The full UNet model with attention and timestep embedding.
    :param in_channels: channels in the input Tensor.
    :param model_channels: base channel count for the model.
    :param out_channels: channels in the output Tensor.
    :param num_res_blocks: number of residual blocks per downsample.
    :param attention_resolutions: a collection of downsample rates at which
        attention will take place. May be a set, list, or tuple.
        For example, if this contains 4, then at 4x downsampling, attention
        will be used.
    :param dropout: the dropout probability.
    :param channel_mult: channel multiplier for each level of the UNet.
    :param conv_resample: if True, use learned convolutions for upsampling and
        downsampling.
    :param dims: determines if the signal is 1D, 2D, or 3D.
    :param num_classes: if specified (as an int), then this model will be
        class-conditional with `num_classes` classes.
    :param use_checkpoint: use gradient checkpointing to reduce memory usage.
    :param num_heads: the number of attention heads in each attention layer.
    :param num_heads_channels: if specified, ignore num_heads and instead use
                               a fixed channel width per attention head.
    :param num_heads_upsample: works with num_heads to set a different number
                               of heads for upsampling. Deprecated.
    :param use_scale_shift_norm: use a FiLM-like conditioning mechanism.
    :param resblock_updown: use residual blocks for up/downsampling.
    """

    def __init__(
        self,
        image_size,
        in_channels,
        model_channels,
        out_channels,
        num_res_blocks,
        attention_resolutions,
        dropout=0,
        channel_mult=(1, 2, 4, 8),
        conv_resample=True,
        dims=2,
        num_classes=None,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=1,
        num_head_channels=-1,
        num_heads_upsample=-1,
        use_scale_shift_norm=False,
        resblock_updown=False,
        use_new_attention_order=False,
        use_spatial_transformer=True,  # custom transformer support
        transformer_depth=1,  # custom transformer support
        context_dim=None,  # custom transformer support
        n_embed=None,  # custom support for prediction of discrete ids into codebook of first stage vq model
        clip_embed_dim=None,  # if set, condition on clip embeddings
        image_condition=False,  # if set, condition on image (for inpainting)
        super_res_condition=False,  # if set, condition on super-resolution data via middle block
    ):
        super().__init__()

        if use_spatial_transformer:
            assert (
                context_dim is not None
            ), "Fool!! You forgot to include the dimension of your cross-attention conditioning..."

        if context_dim is not None:
            assert (
                use_spatial_transformer
            ), "Fool!! You forgot to use the spatial transformer for your cross-attention conditioning..."

        if num_heads_upsample == -1:
            num_heads_upsample = num_heads

        if image_condition:
            in_channels *= 2

        self.image_size = image_size
        self.in_channels = in_channels
        self.model_channels = model_channels
        self.out_channels = out_channels
        self.num_res_blocks = num_res_blocks
        self.attention_resolutions = attention_resolutions
        self.dropout = dropout
        self.channel_mult = channel_mult
        self.conv_resample = conv_resample
        self.num_classes = num_classes
        self.use_checkpoint = use_checkpoint
        self.dtype = th.float16 if use_fp16 else th.float32
        self.num_heads = num_heads
        self.num_head_channels = num_head_channels
        self.num_heads_upsample = num_heads_upsample

        self.predict_codebook_ids = n_embed is not None

        self.use_fp16 = use_fp16
        self.clip_embed_dim = clip_embed_dim
        self.image_condition = image_condition
        self.super_res_condition = super_res_condition

        time_embed_dim = model_channels * 4
        self.time_embed = nn.Sequential(
            linear(model_channels, time_embed_dim),
            nn.SiLU(),
            linear(time_embed_dim, time_embed_dim),
        )

        if clip_embed_dim is not None:
            self.clip_proj = nn.Linear(clip_embed_dim, time_embed_dim, dtype=self.dtype)

        if self.num_classes is not None:
            self.label_emb = nn.Embedding(num_classes, time_embed_dim)

        self.input_blocks = nn.ModuleList(
            [
                TimestepEmbedSequential(
                    conv_nd(dims, in_channels, model_channels, 3, padding=1)
                )
            ]
        )

        self._feature_size = model_channels

        input_block_chans = [model_channels]
        ch = model_channels
        ds = 1
        for level, mult in enumerate(channel_mult):
            for _ in range(num_res_blocks):
                layers = [
                    ResBlock(
                        ch,
                        time_embed_dim,
                        dropout,
                        out_channels=mult * model_channels,
                        dims=dims,
                        use_checkpoint=use_checkpoint,
                        use_scale_shift_norm=use_scale_shift_norm,
                    )
                ]
                ch = mult * model_channels
                if ds in attention_resolutions:
                    dim_head = ch // num_heads
                    layers.append(
                        AttentionBlock(
                            ch,
                            use_checkpoint=use_checkpoint,
                            num_heads=num_heads,
                            num_head_channels=num_head_channels,
                            use_new_attention_order=use_new_attention_order,
                        )
                        if not use_spatial_transformer
                        else SpatialTransformer(
                            ch,
                            num_heads,
                            dim_head,
                            depth=transformer_depth,
                            context_dim=context_dim,
                        )
                    )
                self.input_blocks.append(TimestepEmbedSequential(*layers))
                self._feature_size += ch
                input_block_chans.append(ch)
            if level != len(channel_mult) - 1:
                out_ch = ch
                self.input_blocks.append(
                    TimestepEmbedSequential(
                        ResBlock(
                            ch,
                            time_embed_dim,
                            dropout,
                            out_channels=out_ch,
                            dims=dims,
                            use_checkpoint=use_checkpoint,
                            use_scale_shift_norm=use_scale_shift_norm,
                            down=True,
                        )
                        if resblock_updown
                        else Downsample(
                            ch, conv_resample, dims=dims, out_channels=out_ch
                        )
                    )
                )
                ch = out_ch
                input_block_chans.append(ch)
                ds *= 2
                self._feature_size += ch

        if self.super_res_condition:
            self.external_block = nn.Sequential(
                Downsample(4, True, out_channels=64),
                Downsample(64, True, out_channels=128),
                Downsample(256, True, out_channels=256),
            )
            middle_ch = ch + 256
        else:
            middle_ch = ch

        dim_head = ch // num_heads
        self.middle_block = TimestepEmbedSequential(
            ResBlock(
                ch,
                time_embed_dim,
                dropout,
                dims=dims,
                use_checkpoint=use_checkpoint,
                use_scale_shift_norm=use_scale_shift_norm,
            ),
            AttentionBlock(
                ch,
                use_checkpoint=use_checkpoint,
                num_heads=num_heads,
                num_head_channels=num_head_channels,
                use_new_attention_order=use_new_attention_order,
            )
            if not use_spatial_transformer
            else SpatialTransformer(
                ch,
                num_heads,
                dim_head,
                depth=transformer_depth,
                context_dim=context_dim,
            ),
            ResBlock(
                ch,
                time_embed_dim,
                dropout,
                dims=dims,
                use_checkpoint=use_checkpoint,
                use_scale_shift_norm=use_scale_shift_norm,
            ),
        )
        self._feature_size += ch

        # if self.emb_condition_channels > 0:
        #    self.middle_block_cond = middle_block
        # else:
        #    self.middle_block = middle_block

        self.output_blocks = nn.ModuleList([])
        for level, mult in list(enumerate(channel_mult))[::-1]:
            for i in range(num_res_blocks + 1):
                ich = input_block_chans.pop()
                layers = [
                    ResBlock(
                        ch + ich,
                        time_embed_dim,
                        dropout,
                        out_channels=model_channels * mult,
                        dims=dims,
                        use_checkpoint=use_checkpoint,
                        use_scale_shift_norm=use_scale_shift_norm,
                    )
                ]
                ch = model_channels * mult
                if ds in attention_resolutions:
                    dim_head = ch // num_heads
                    layers.append(
                        AttentionBlock(
                            ch,
                            use_checkpoint=use_checkpoint,
                            num_heads=num_heads_upsample,
                            num_head_channels=num_head_channels,
                            use_new_attention_order=use_new_attention_order,
                        )
                        if not use_spatial_transformer
                        else SpatialTransformer(
                            ch,
                            num_heads,
                            dim_head,
                            depth=transformer_depth,
                            context_dim=context_dim,
                        )
                    )
                if level and i == num_res_blocks:
                    out_ch = ch
                    layers.append(
                        ResBlock(
                            ch,
                            time_embed_dim,
                            dropout,
                            out_channels=out_ch,
                            dims=dims,
                            use_checkpoint=use_checkpoint,
                            use_scale_shift_norm=use_scale_shift_norm,
                            up=True,
                        )
                        if resblock_updown
                        else Upsample(ch, conv_resample, dims=dims, out_channels=out_ch)
                    )
                    ds //= 2
                self.output_blocks.append(TimestepEmbedSequential(*layers))
                self._feature_size += ch

        self.out = nn.Sequential(
            normalization(ch),
            nn.SiLU(),
            zero_module(conv_nd(dims, model_channels, out_channels, 3, padding=1)),
        )
        if self.predict_codebook_ids:
            self.id_predictor = nn.Sequential(
                normalization(ch),
                conv_nd(dims, model_channels, n_embed, 1),
                # nn.LogSoftmax(dim=1)  # change to cross_entropy and produce non-normalized logits
            )

    def convert_to_fp16(self):
        """
        Convert the torso of the model to float16.
        """
        if self.clip_embed_dim is not None:
            self.clip_proj.weight.data = self.clip_proj.weight.data.half()
            if self.clip_proj.bias is not None:
                self.clip_proj.bias.data = self.clip_proj.bias.data.half()

        self.input_blocks.apply(convert_module_to_f16)

        if self.super_res_condition:
            self.external_block.apply(convert_module_to_f16)
        #    self.middle_block_cond.apply(convert_module_to_f16)
        # else:
        self.middle_block.apply(convert_module_to_f16)

        self.output_blocks.apply(convert_module_to_f16)

    def convert_to_fp32(self):
        """
        Convert the torso of the model to float32.
        """
        if self.clip_embed_dim is not None:
            self.clip_proj.weight.data = self.clip_proj.weight.data.float()
            if self.clip_proj.bias is not None:
                self.clip_proj.bias.data = self.clip_proj.bias.data.float()

        self.input_blocks.apply(convert_module_to_f32)

        if self.super_res_condition:
            self.external_block.apply(convert_module_to_f32)
            # self.middle_block_cond.apply(convert_module_to_f32)
        # else:
        self.middle_block.apply(convert_module_to_f32)

        self.output_blocks.apply(convert_module_to_f32)

    def forward(
        self,
        x,
        timesteps=None,
        context=None,
        clip_embed=None,
        image_embed=None,
        super_res_embed=None,
        y=None,
        **kwargs,
    ):
        """
        Apply the model to an input batch.
        :param x: an [N x C x ...] Tensor of inputs.
        :param timesteps: a 1-D batch of timesteps.
        :param context: conditioning plugged in via crossattn
        :param clip_embed: a clip text or image embedding
        :param image_embed: image embeddings to be concatenated with x (for inpainting)
        :param super_res_embed: image embeddings to be fed into the middle block (for upscaling)
        :param y: an [N] Tensor of labels, if class-conditional.
        :return: an [N x C x ...] Tensor of outputs.
        """
        assert (y is not None) == (
            self.num_classes is not None
        ), "must specify y if and only if the model is class-conditional"
        assert timesteps is not None, "need to implement no-timestep usage"
        hs = []
        t_emb = timestep_embedding(timesteps, self.model_channels, repeat_only=False)
        emb = self.time_embed(t_emb)

        if clip_embed is not None:
            emb = emb + self.clip_proj(clip_embed).to(emb)

        if self.num_classes is not None:
            assert y.shape == (x.shape[0],)
            emb = emb + self.label_emb(y)

        if image_embed is not None:
            x = th.cat([x, image_embed], dim=1)

        h = x.type(self.dtype)

        for module in self.input_blocks:
            h = module(h, emb, context)
            hs.append(h)

        if super_res_embed is not None:
            im = super_res_embed.type(self.dtype)
            im = self.external_block(im)
            h = th.cat([h, im], dim=1)

        h = self.middle_block(h, emb, context)
        for module in self.output_blocks:
            h = th.cat([h, hs.pop()], dim=1)
            h = module(h, emb, context)
        h = h.type(x.dtype)
        if self.predict_codebook_ids:
            # return self.out(h), self.id_predictor(h)
            return self.id_predictor(h)
        else:
            return self.out(h)



from pathlib import Path
from omegaconf import OmegaConf
import torch
import torch
from basicsr.utils.download_util import load_file_from_url

from .ldm.util import instantiate_from_config
from perceptor.utils import cache
from . import diffusion_space

CONFIG_DIR = Path(__file__).resolve().parent / "configs"


@cache
class Face(torch.nn.Module):
    def __init__(self, eta=0.0):
        """
        Usage:
            from tqdm import tqdm
            import perceptor

            model = perceptor.models.latent_diffusion.Face().cuda()
            diffused_latents = model.random_latents((1, 3, 256, 256)).cuda()

            for from_index, to_index in tqdm(model.schedule_indices(n_steps=50)):
                denoised_latents = model.denoise(diffused_latents, from_index)
                diffused_latents = model.step(
                    diffused_latents, denoised_latents, from_index, to_index
                )
            denoised_latents = model.denoise(diffused_latents, to_index)
            images = model.images(denoised_latents)
        """
        super().__init__()
        self.eta = eta

        autoencoder_url_ckpt = "https://s3.eu-central-1.wasabisys.com/nextml-model-data/latent-diffusion/celeba.pt"
        load_file_from_url(
            autoencoder_url_ckpt, "models", file_name="latent-diffusion-vq-f4.pt"
        )

        url_ckpt = "https://s3.eu-central-1.wasabisys.com/nextml-model-data/latent-diffusion/celeba.pt"

        config_path = CONFIG_DIR / "latent-diffusion" / "celebahq-ldm-vq-4.yaml"
        checkpoint_path = load_file_from_url(
            url_ckpt, "models", file_name="latent-diffusion-celeba.pt"
        )

        config = OmegaConf.load(config_path)
        self.model = instantiate_from_config(config.model)
        self.model.load_state_dict(
            torch.load(checkpoint_path, map_location="cpu")["state_dict"], strict=False
        )
        self.model.requires_grad_(False)
        self.model.cuda()
        self.model.eval()

    @property
    def device(self):
        return next(iter(self.parameters())).device

    def schedule_indices(self, from_index=999, to_index=50, n_steps=None):
        if from_index < to_index:
            raise ValueError("from_index must be greater than to_index")
        if n_steps is None:
            n_steps = (from_index - to_index) // 2
        schedule_indices = torch.linspace(from_index, to_index, n_steps).long()
        from_indices = schedule_indices[:-1]
        to_indices = schedule_indices[1:]
        if (from_indices == to_indices).any():
            raise ValueError("Schedule indices must be unique")
        return torch.stack([from_indices, to_indices], dim=1)

    @staticmethod
    def latent_shape(height, width):
        return [3, height // 4, width // 4]

    def forward(self, latents, index):
        return self.velocity(latents, index)

    def random_latents(self, images_shape):
        return torch.randn(
            (images_shape[0], *self.latent_shape(*images_shape[-2:])),
            device=self.device,
        )

    def latents(self, images):
        if images.shape[-2:] != (256, 256):
            raise ValueError(
                "Face model not behave well when resolution is not 256x256"
            )
        encoder_posterior = self.model.encode_first_stage(
            diffusion_space.encode(images)
        )
        return self.model.get_first_stage_encoding(encoder_posterior)

    def diffuse(self, latents, index, noise=None):
        """Unclear what the first argument should be. Conditioning works and latents also works okay"""
        if noise is None:
            noise = torch.randn_like(latents)
        else:
            if latents.shape != noise.shape:
                raise ValueError("Noise shape must be the same as latent shape")

        return self.model.q_sample(x_start=latents, t=self.ts(index), noise=noise)

    def denoise(self, latents, index, eps=None):
        """Predict denoised latents"""
        if index >= 1000:
            raise ValueError("index must be less than 1000")

        if eps is None:
            eps = self.eps(latents, index)

        return (
            latents - self.sqrt_one_minus_alphas_cumprod(index) * eps
        ) / self.alphas_cumprod(index).sqrt()

    def images(self, latents):
        """Decode from latent space to image"""
        return diffusion_space.decode(self.model.decode_first_stage(latents))

    def ts(self, index):
        return torch.tensor([index], device=self.device, dtype=torch.long)

    def alphas_cumprod(self, index):
        return self.model.alphas_cumprod[index].to(self.device)[None, None, None, None]

    def sqrt_one_minus_alphas_cumprod(self, index):
        return self.model.sqrt_one_minus_alphas_cumprod[index].to(self.device)[
            None, None, None, None
        ]

    def step(
        self, from_latents, predicted_denoised_latents, from_index, to_index, noise=None
    ):
        if to_index > from_index:
            raise ValueError("to_index must be smaller than from_index")
        if noise is None:
            noise = torch.randn_like(predicted_denoised_latents)
        else:
            if from_latents.shape != noise.shape:
                raise ValueError("Noise shape must be the same as latent shape")

        # if quantize_denoised:
        #     predicted_denoised_latents, _, *_ = self.model.first_stage_model.quantize(predicted_denoised_latents)

        from_alphas_cumprod = self.alphas_cumprod(from_index)
        to_alphas_cumprod = self.alphas_cumprod(to_index)
        from_sqrt_one_minus_alphas_cumprod = self.sqrt_one_minus_alphas_cumprod(
            from_index
        )

        to_sigmas = self.eta * torch.sqrt(
            (1 - to_alphas_cumprod)
            / (1 - from_alphas_cumprod)
            * (1 - from_alphas_cumprod / to_alphas_cumprod)
        )

        eps = (
            from_latents - predicted_denoised_latents * from_alphas_cumprod.sqrt()
        ) / from_sqrt_one_minus_alphas_cumprod

        dir_xt = (1.0 - to_alphas_cumprod - to_sigmas**2).sqrt() * eps

        to_z = (
            to_alphas_cumprod.sqrt() * predicted_denoised_latents
            + dir_xt
            + to_sigmas * noise
        )
        return to_z

    def eps(self, latents, index):
        return self.model.apply_model(latents, self.ts(index), cond=None)


def test_face():
    import perceptor

    model = perceptor.models.latent_diffusion.Face().cuda()
    diffused_latents = model.random_latents((1, 3, 256, 256)).cuda()

    for from_index, to_index in model.schedule_indices(to_index=50, n_steps=50):
        denoised_latents = model.denoise(diffused_latents, from_index)
        diffused_latents = model.step(
            diffused_latents, denoised_latents, from_index, to_index
        )
    denoised_latents = model.denoise(diffused_latents, to_index)
    images = model.images(denoised_latents)
    perceptor.utils.pil_image(images).save("tests/face.png")



# -*- coding: utf-8 -*-
import os
import json
from collections import OrderedDict

import torch
import numpy as np
from torch import nn


class LayerNorm(nn.LayerNorm):
    """Subclass torch's LayerNorm to handle fp16."""

    def forward(self, x: torch.Tensor):
        orig_type = x.dtype
        ret = super().forward(x.type(torch.float32))
        return ret.type(orig_type)


class QuickGELU(nn.Module):
    def forward(self, x: torch.Tensor):
        return x * torch.sigmoid(1.702 * x)


class ResidualAttentionBlock(nn.Module):
    def __init__(self, d_model: int, n_head: int, attn_mask: torch.Tensor = None):
        super().__init__()

        self.attn = nn.MultiheadAttention(d_model, n_head)
        self.ln_1 = LayerNorm(d_model)
        self.mlp = nn.Sequential(
            OrderedDict(
                [
                    ("c_fc", nn.Linear(d_model, d_model * 4)),
                    ("gelu", QuickGELU()),
                    ("c_proj", nn.Linear(d_model * 4, d_model)),
                ]
            )
        )
        self.ln_2 = LayerNorm(d_model)
        self.attn_mask = attn_mask

    def attention(self, x: torch.Tensor):
        self.attn_mask = (
            self.attn_mask.to(dtype=x.dtype, device=x.device)
            if self.attn_mask is not None
            else None
        )
        return self.attn(x, x, x, need_weights=False, attn_mask=self.attn_mask)[0]

    def forward(self, x: torch.Tensor):
        x = x + self.attention(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


class Transformer(nn.Module):
    def __init__(
        self, width: int, layers: int, heads: int, attn_mask: torch.Tensor = None
    ):
        super().__init__()
        self.width = width
        self.layers = layers
        self.resblocks = nn.Sequential(
            *[ResidualAttentionBlock(width, heads, attn_mask) for _ in range(layers)]
        )

    def forward(self, x: torch.Tensor):
        return self.resblocks(x)


class VisionTransformer(nn.Module):
    def __init__(
        self,
        input_resolution: int,
        patch_size: int,
        width: int,
        layers: int,
        heads: int,
        output_dim: int,
    ):
        super().__init__()
        self.input_resolution = input_resolution
        self.output_dim = output_dim
        self.conv1 = nn.Conv2d(
            in_channels=3,
            out_channels=width,
            kernel_size=patch_size,
            stride=patch_size,
            bias=False,
        )

        scale = width**-0.5
        self.class_embedding = nn.Parameter(scale * torch.randn(width))
        self.positional_embedding = nn.Parameter(
            scale * torch.randn((input_resolution // patch_size) ** 2 + 1, width)
        )
        self.ln_pre = LayerNorm(width)

        self.transformer = Transformer(width, layers, heads)

        self.ln_post = LayerNorm(width)
        self.proj = nn.Parameter(scale * torch.randn(width, output_dim))

    def forward(self, x: torch.Tensor):
        x = self.conv1(x)  # shape = [*, width, grid, grid]
        x = x.reshape(x.shape[0], x.shape[1], -1)  # shape = [*, width, grid ** 2]
        x = x.permute(0, 2, 1)  # shape = [*, grid ** 2, width]
        x = torch.cat(
            [
                self.class_embedding.to(x.dtype)
                + torch.zeros(
                    x.shape[0], 1, x.shape[-1], dtype=x.dtype, device=x.device
                ),
                x,
            ],
            dim=1,
        )  # shape = [*, grid ** 2 + 1, width]
        x = x + self.positional_embedding.to(x.dtype)
        x = self.ln_pre(x)

        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD

        x = self.ln_post(x[:, 0, :])

        if self.proj is not None:
            x = x @ self.proj

        return x


class CLIP(nn.Module):
    def __init__(
        self,
        embed_dim,
        image_resolution,
        vision_layers,
        vision_width,
        vision_patch_size,
        context_length,
        vocab_size,
        transformer_width,
        transformer_heads,
        transformer_layers,
        eos_id=3,
    ):
        super().__init__()

        self.eos_id = eos_id
        self.context_length = context_length

        vision_heads = vision_width // 64
        self.visual = VisionTransformer(
            input_resolution=image_resolution,
            patch_size=vision_patch_size,
            width=vision_width,
            layers=vision_layers,
            heads=vision_heads,
            output_dim=embed_dim,
        )

        self.transformer = Transformer(
            width=transformer_width,
            layers=transformer_layers,
            heads=transformer_heads,
            attn_mask=self.build_attention_mask(),
        )

        self.vocab_size = vocab_size
        self.token_embedding = nn.Embedding(vocab_size, transformer_width)
        self.positional_embedding = nn.Parameter(
            torch.empty(self.context_length, transformer_width)
        )
        self.ln_final = LayerNorm(transformer_width)

        self.text_projection = nn.Parameter(torch.empty(transformer_width, embed_dim))
        self.logit_scale = nn.Parameter(torch.ones([]) * np.log(1 / 0.07))

    def build_attention_mask(self):
        mask = torch.empty(self.context_length, self.context_length)
        mask.fill_(float("-inf"))
        mask.triu_(1)
        return mask

    @property
    def dtype(self):
        return self.visual.conv1.weight.dtype

    def encode_image(self, pixel_values):
        """Encode images
        Parameters
        ----------
        pixel_values: torch.Tensor
            Processed images from RuCLIPProcessor class
        Returns
        -------
        image_latents : torch.Tensor
            Image embeddings
        """
        return self.visual(pixel_values.type(self.dtype))

    def encode_text(self, input_ids):
        """Encode texts
        Parameters
        ----------
        input_ids: torch.Tensor
            Tokenized texts from RuCLIPProcessor class
        Returns
        -------
        text_latents : torch.Tensor
            Text embeddings
        """
        x = self.token_embedding(input_ids).type(
            self.dtype
        )  # [batch_size, n_ctx, d_model]
        x = x + self.positional_embedding.type(self.dtype)
        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD
        x = self.ln_final(x).type(self.dtype)
        # x.shape = [batch_size, n_ctx, transformer.width]
        x = (
            x[torch.arange(x.shape[0]), torch.where(input_ids == self.eos_id)[1]]
            @ self.text_projection
        )
        return x

    def forward(self, input_ids, pixel_values):
        image_features = self.encode_image(pixel_values)
        text_features = self.encode_text(input_ids)

        # normalize features
        image_features = image_features / image_features.norm(dim=-1, keepdim=True)
        text_features = text_features / text_features.norm(dim=-1, keepdim=True)

        # cosine similarity as logits
        logit_scale = self.logit_scale.exp()
        logits_per_image = logit_scale * image_features @ text_features.t()
        logits_per_text = logits_per_image.t()

        return logits_per_image, logits_per_text

    @classmethod
    def from_pretrained(cls, folder):
        """Load model from folder"""
        config = json.load(open(os.path.join(folder, "config.json")))
        model = cls(
            embed_dim=config["embed_dim"],
            image_resolution=config["image_resolution"],
            vision_layers=config["vision_layers"],
            vision_width=config["vision_width"],
            vision_patch_size=config["vision_patch_size"],
            context_length=config["context_length"],
            vocab_size=config["vocab_size"],
            transformer_width=config["transformer_width"],
            transformer_heads=config["transformer_heads"],
            transformer_layers=config["transformer_layers"],
        )
        checkpoint = torch.load(
            os.path.join(folder, "pytorch_model.bin"), map_location="cpu"
        )
        model.load_state_dict(checkpoint)
        return model



import torch
import torch.nn.functional as F

from perceptor import utils
from perceptor.transforms.resize import resize
from . import model_pt, pretrained


models = {
    "16-epochs": "cloob_laion_400m_vit_b_16_16_epochs",
    "32-epochs": "cloob_laion_400m_vit_b_16_32_epochs",
}


@utils.cache
class CLOOB(torch.nn.Module):
    def __init__(self, name="16-epochs"):
        """
        Args:
            name: name of the cloob model. Available models are:
                - 16-epochs
                - 32-epochs
        """
        super().__init__()
        self.name = name
        config = pretrained.get_config(models[self.name])
        image_size = config["image_encoder"]["image_size"]
        self.image_size = (image_size, image_size)
        self.model = model_pt.get_pt_model(config)
        checkpoint = pretrained.download_checkpoint(config)
        self.model.load_state_dict(model_pt.get_pt_params(config, checkpoint))
        self.model.eval().requires_grad_(False)

    @property
    def device(self):
        return next(iter(self.parameters()))

    def encode_texts(self, text_prompts):
        return self.model.text_encoder(self.model.tokenize(text_prompts))

    def encode_images(self, images):
        if tuple(images.shape[-2:]) != self.image_size:
            images = resize(
                images,
                out_shape=self.image_size,
            )
        return F.normalize(
            self.model.image_encoder(
                self.model.normalize(images.to(self.device))
            ).float(),
            dim=-1,
        )

    def forward(self, _):
        raise NotImplementedError




def append_dims(x, target_dims):
    """Appends dimensions to the end of a tensor until it has target_dims dimensions."""
    dims_to_append = target_dims - x.ndim
    if dims_to_append < 0:
        raise ValueError(
            f"input has {x.ndim} dims but target_dims is {target_dims}, which is less"
        )
    return x[(...,) + (None,) * dims_to_append]



# -*- coding: utf-8 -*-
import numpy as np


def pad_reflect(image, pad_size):
    imsize = image.shape
    height, width = imsize[:2]
    new_img = np.zeros([height + pad_size * 2, width + pad_size * 2, imsize[2]]).astype(
        np.uint8
    )
    new_img[pad_size:-pad_size, pad_size:-pad_size, :] = image
    new_img[0:pad_size, pad_size:-pad_size, :] = np.flip(
        image[0:pad_size, :, :], axis=0
    )  # top
    new_img[-pad_size:, pad_size:-pad_size, :] = np.flip(
        image[-pad_size:, :, :], axis=0
    )  # bottom
    new_img[:, 0:pad_size, :] = np.flip(
        new_img[:, pad_size : pad_size * 2, :], axis=1
    )  # left
    new_img[:, -pad_size:, :] = np.flip(
        new_img[:, -pad_size * 2 : -pad_size, :], axis=1
    )  # right
    return new_img


def unpad_image(image, pad_size):
    return image[pad_size:-pad_size, pad_size:-pad_size, :]


def pad_patch(image_patch, padding_size, channel_last=True):
    """Pads image_patch with with padding_size edge values."""
    if channel_last:
        return np.pad(
            image_patch,
            ((padding_size, padding_size), (padding_size, padding_size), (0, 0)),
            "edge",
        )
    else:
        return np.pad(
            image_patch,
            ((0, 0), (padding_size, padding_size), (padding_size, padding_size)),
            "edge",
        )


def unpad_patches(image_patches, padding_size):
    return image_patches[:, padding_size:-padding_size, padding_size:-padding_size, :]


def split_image_into_overlapping_patches(image_array, patch_size, padding_size=2):
    """Splits the image into partially overlapping patches.
    The patches overlap by padding_size pixels.
    Pads the image twice:
        - first to have a size multiple of the patch size,
        - then to have equal padding at the borders.
    Args:
        image_array: numpy array of the input image.
        patch_size: size of the patches from the original image (without padding).
        padding_size: size of the overlapping area.
    """
    xmax, ymax, _ = image_array.shape
    x_remainder = xmax % patch_size
    y_remainder = ymax % patch_size

    # modulo here is to avoid extending of patch_size instead of 0
    x_extend = (patch_size - x_remainder) % patch_size
    y_extend = (patch_size - y_remainder) % patch_size

    # make sure the image is divisible into regular patches
    extended_image = np.pad(image_array, ((0, x_extend), (0, y_extend), (0, 0)), "edge")

    # add padding around the image to simplify computations
    padded_image = pad_patch(extended_image, padding_size, channel_last=True)

    xmax, ymax, _ = padded_image.shape
    patches = []

    x_lefts = range(padding_size, xmax - padding_size, patch_size)
    y_tops = range(padding_size, ymax - padding_size, patch_size)

    for x in x_lefts:
        for y in y_tops:
            x_left = x - padding_size
            y_top = y - padding_size
            x_right = x + patch_size + padding_size
            y_bottom = y + patch_size + padding_size
            patch = padded_image[x_left:x_right, y_top:y_bottom, :]
            patches.append(patch)

    return np.array(patches), padded_image.shape


def stich_together(patches, padded_image_shape, target_shape, padding_size=4):
    """Reconstruct the image from overlapping patches.
    After scaling, shapes and padding should be scaled too.
    Args:
        patches: patches obtained with split_image_into_overlapping_patches
        padded_image_shape: shape of the padded image contructed in split_image_into_overlapping_patches
        target_shape: shape of the final image
        padding_size: size of the overlapping area.
    """

    xmax, ymax, _ = padded_image_shape
    patches = unpad_patches(patches, padding_size)
    patch_size = patches.shape[1]
    n_patches_per_row = ymax // patch_size

    complete_image = np.zeros((xmax, ymax, 3))

    row = -1
    col = 0
    for i in range(len(patches)):
        if i % n_patches_per_row == 0:
            row += 1
            col = 0
        complete_image[
            row * patch_size : (row + 1) * patch_size,
            col * patch_size : (col + 1) * patch_size,
            :,
        ] = patches[i]
        col += 1
    return complete_image[0 : target_shape[0], 0 : target_shape[1], :]



# -*- coding: utf-8 -*-
import os
import json

import torch
import numpy as np
import youtokentome as yttm
import torchvision.transforms as T
from torch.nn.utils.rnn import pad_sequence


class RuCLIPProcessor(torch.nn.Module):
    eos_id = 3
    bos_id = 2
    unk_id = 1
    pad_id = 0

    def __init__(
        self, tokenizer_path, image_size=224, text_seq_length=77, mean=None, std=None
    ):
        super().__init__()
        self.tokenizer = yttm.BPE(tokenizer_path)
        self.mean = mean or [0.48145466, 0.4578275, 0.40821073]
        self.std = std or [0.26862954, 0.26130258, 0.27577711]
        self.image_transform = T.Compose(
            [
                T.Resize(image_size),
                T.Normalize(mean=self.mean, std=self.std),
            ]
        )
        self.text_seq_length = text_seq_length
        self.image_size = image_size

    def encode_text(self, text):
        text = text.lower()
        tokens = self.tokenizer.encode(
            [text], output_type=yttm.OutputType.ID, dropout_prob=0.0
        )[0]
        tokens = [self.bos_id] + tokens + [self.eos_id]
        tokens = tokens[: self.text_seq_length]
        return self.prepare_tokens(tokens)

    def prepare_tokens(self, tokens):
        empty_positions = self.text_seq_length - len(tokens)
        if empty_positions > 0:
            tokens = np.hstack(
                (tokens, np.zeros(empty_positions))
            )  # position tokens after text
        if len(tokens) > self.text_seq_length:
            tokens = tokens[: self.text_seq_length - 1] + tokens[-1:]
        return torch.tensor(tokens).long()

    def decode_text(self, encoded):
        return self.tokenizer.decode(
            encoded.cpu().numpy().tolist(),
            ignore_ids=[self.eos_id, self.bos_id, self.unk_id, self.pad_id],
        )[0]

    def __call__(self, text=None, images=None, **kwargs):
        inputs = {}
        if text is not None:
            input_ids = []
            texts = [text] if isinstance(text, str) else text
            for text in texts:
                tokens = self.encode_text(text)
                input_ids.append(tokens)
            inputs["input_ids"] = pad_sequence(input_ids, batch_first=True)
        if images is not None:
            pixel_values = []
            for i, image in enumerate(images):
                pixel_values.append(self.image_transform(image))
            inputs["pixel_values"] = pad_sequence(pixel_values, batch_first=True)
        return inputs

    @classmethod
    def from_pretrained(cls, folder):
        tokenizer_path = os.path.join(folder, "bpe.model")
        config = json.load(open(os.path.join(folder, "config.json")))
        image_size = config["image_resolution"]
        text_seq_length = config["context_length"]
        mean, std = config.get("mean"), config.get("std")
        return cls(
            tokenizer_path,
            image_size=image_size,
            text_seq_length=text_seq_length,
            mean=mean,
            std=std,
        )



from torch import nn


class TransformInterface(nn.Module):
    def forward(self, *args, **kwargs):
        return self.encode(*args, **kwargs)

    def encode(self, *args, **kwargs):
        raise NotImplementedError

    def decode(self, *args, **kwargs):
        raise NotImplementedError



"""
Various utilities for neural networks.
"""

import math

import torch as th
import torch.nn as nn


# PyTorch 1.7 has SiLU, but we support PyTorch 1.5.
class SiLU(nn.Module):
    def forward(self, x):
        return x * th.sigmoid(x)


class GroupNorm32(nn.GroupNorm):
    def forward(self, x):
        return super().forward(x.float()).type(x.dtype)


def conv_nd(dims, *args, **kwargs):
    """
    Create a 1D, 2D, or 3D convolution module.
    """
    if dims == 1:
        return nn.Conv1d(*args, **kwargs)
    elif dims == 2:
        return nn.Conv2d(*args, **kwargs)
    elif dims == 3:
        return nn.Conv3d(*args, **kwargs)
    raise ValueError(f"unsupported dimensions: {dims}")


def linear(*args, **kwargs):
    """
    Create a linear module.
    """
    return nn.Linear(*args, **kwargs)


def avg_pool_nd(dims, *args, **kwargs):
    """
    Create a 1D, 2D, or 3D average pooling module.
    """
    if dims == 1:
        return nn.AvgPool1d(*args, **kwargs)
    elif dims == 2:
        return nn.AvgPool2d(*args, **kwargs)
    elif dims == 3:
        return nn.AvgPool3d(*args, **kwargs)
    raise ValueError(f"unsupported dimensions: {dims}")


def update_ema(target_params, source_params, rate=0.99):
    """
    Update target parameters to be closer to those of source parameters using
    an exponential moving average.
    :param target_params: the target parameter sequence.
    :param source_params: the source parameter sequence.
    :param rate: the EMA rate (closer to 1 means slower).
    """
    for targ, src in zip(target_params, source_params):
        targ.detach().mul_(rate).add_(src, alpha=1 - rate)


def zero_module(module):
    """
    Zero out the parameters of a module and return it.
    """
    for p in module.parameters():
        p.detach().zero_()
    return module


def scale_module(module, scale):
    """
    Scale the parameters of a module and return it.
    """
    for p in module.parameters():
        p.detach().mul_(scale)
    return module


def mean_flat(tensor):
    """
    Take the mean over all non-batch dimensions.
    """
    return tensor.mean(dim=list(range(1, len(tensor.shape))))


def normalization(channels):
    """
    Make a standard normalization layer.
    :param channels: number of input channels.
    :return: an nn.Module for normalization.
    """
    return GroupNorm32(32, channels)


def timestep_embedding(timesteps, dim, max_period=10000):
    """
    Create sinusoidal timestep embeddings.
    :param timesteps: a 1-D Tensor of N indices, one per batch element.
                      These may be fractional.
    :param dim: the dimension of the output.
    :param max_period: controls the minimum frequency of the embeddings.
    :return: an [N x dim] Tensor of positional embeddings.
    """
    half = dim // 2
    freqs = th.exp(
        -math.log(max_period) * th.arange(start=0, end=half, dtype=th.float32) / half
    ).to(device=timesteps.device)
    args = timesteps[:, None].float() * freqs[None]
    embedding = th.cat([th.cos(args), th.sin(args)], dim=-1)
    if dim % 2:
        embedding = th.cat([embedding, th.zeros_like(embedding[:, :1])], dim=-1)
    return embedding


def checkpoint(func, inputs, params, flag):
    """
    Evaluate a function without caching intermediate activations, allowing for
    reduced memory at the expense of extra compute in the backward pass.
    :param func: the function to evaluate.
    :param inputs: the argument sequence to pass to `func`.
    :param params: a sequence of parameters `func` depends on but does not
                   explicitly take as arguments.
    :param flag: if False, disable gradient checkpointing.
    """
    if flag:
        args = tuple(inputs) + tuple(params)
        return CheckpointFunction.apply(func, len(inputs), *args)
    else:
        return func(*inputs)


class CheckpointFunction(th.autograd.Function):
    @staticmethod
    @th.cuda.amp.custom_fwd
    def forward(ctx, run_function, length, *args):
        ctx.run_function = run_function
        ctx.input_length = length
        ctx.save_for_backward(*args)
        with th.no_grad():
            output_tensors = ctx.run_function(*args[:length])
        return output_tensors

    @staticmethod
    @th.cuda.amp.custom_bwd
    def backward(ctx, *output_grads):
        args = list(ctx.saved_tensors)

        # Filter for inputs that require grad. If none, exit early.
        input_indices = [i for (i, x) in enumerate(args) if x.requires_grad]
        if not input_indices:
            return (None, None) + tuple(None for _ in args)

        with th.enable_grad():
            for i in input_indices:
                if i < ctx.input_length:
                    # Not sure why the OAI code does this little
                    # dance. It might not be necessary.
                    args[i] = args[i].detach().requires_grad_()
                    args[i] = args[i].view_as(args[i])
            output_tensors = ctx.run_function(*args[: ctx.input_length])

        if isinstance(output_tensors, th.Tensor):
            output_tensors = [output_tensors]

        # Filter for outputs that require grad. If none, exit early.
        out_and_grads = [
            (o, g) for (o, g) in zip(output_tensors, output_grads) if o.requires_grad
        ]
        if not out_and_grads:
            return (None, None) + tuple(None for _ in args)

        # Compute gradients on the filtered tensors.
        computed_grads = th.autograd.grad(
            [o for (o, g) in out_and_grads],
            [args[i] for i in input_indices],
            [g for (o, g) in out_and_grads],
        )

        # Reassemble the complete gradient tuple.
        input_grads = [None for _ in args]
        for (i, g) in zip(input_indices, computed_grads):
            input_grads[i] = g
        return (None, None) + tuple(input_grads)



from PIL import Image
from torchvision.transforms.functional import to_pil_image
from lantern import Tensor


def pil_image(images: Tensor) -> Image:
    if images.max() > 1 or images.min() < 0:
        print("Warning: images are not in range [0, 1]")
    n, c, h, w = images.shape
    return to_pil_image(
        images.permute(0, 2, 3, 1).reshape(-1, w, c).permute(2, 0, 1).clamp(0, 1)
    )



from PIL import Image
import numpy as np
import torch
import torch.nn.functional as F
from lantern import FunctionalBase, Tensor

from . import standardize, diffusion, settings


class Prediction(FunctionalBase):
    denoised_image: Tensor.dims("CHW").shape(*settings.INPUT_SHAPE)
    diffused_image: Tensor.dims("CHW").shape(*settings.INPUT_SHAPE)
    t: Tensor.dims("").float()

    def representation(self, example):
        denoised_image = self.denoised_image.detach().clamp(0, 1) * 255
        diffused_image = self.diffused_image.detach().clamp(0, 1) * 255
        horizontal_line = torch.full((3, 2, denoised_image.shape[-1]), 255)
        return np.uint8(
            torch.cat(
                [
                    torch.from_numpy(example.image).clamp(0, 255).permute(2, 0, 1),
                    horizontal_line,
                    diffused_image,
                    horizontal_line,
                    denoised_image,
                ],
                dim=-2,
            )
            .permute(1, 2, 0)
            .numpy()
        )

    def pil_image(self, example):
        return Image.fromarray(self.representation(example))


class PredictionBatch(FunctionalBase):
    denoised_xs: Tensor.dims("NCHW").shape(-1, *settings.INPUT_SHAPE).float()
    diffused_images: Tensor.dims("NCHW").shape(-1, *settings.INPUT_SHAPE).float()
    ts: Tensor.dims("N").float()

    def __len__(self):
        return len(self.denoised_images)

    def __getitem__(self, index):
        return Prediction(
            denoised_image=self.denoised_images[index],
            diffused_image=self.diffused_images[index],
            t=self.ts[index],
        )

    @property
    def device(self):
        return self.denoised_xs.device

    def __iter__(self):
        for index in range(len(self)):
            yield self[index]

    @staticmethod
    def sigmas(ts):
        if isinstance(ts, float):
            ts = torch.as_tensor(ts)
        if ts.ndim == 0:
            return torch.full((1,), ts).to(ts.device)
        return ts[:, None, None, None]

    @staticmethod
    def alphas(ts):
        return torch.ones_like(PredictionBatch.sigmas(ts))

    @property
    def from_sigmas(self):
        return self.sigmas(self.ts).to(self.device)

    @property
    def from_alphas(self):
        return self.alphas(self.ts).to(self.device)

    @property
    def from_xs(self):
        return self.diffused_xs

    @property
    def diffused_xs(self):
        return standardize.encode(self.diffused_images).to(self.device)

    @property
    def denoised_images(self):
        return standardize.decode(self.denoised_xs)

    @property
    def eps(self):
        return (self.diffused_xs - self.denoised_xs) / self.from_sigmas

    def step(self, to_ts):
        """
        Step the diffused image forward to `to_t`. Decreasing the amount of noise
        by moving closer to the predicted denoised image.
        """
        to_ts = to_ts.to(self.device)
        to_alphas, to_sigmas = self.alphas(to_ts), self.sigmas(to_ts)

        to_diffused_xs = self.denoised_xs * to_alphas + self.eps * to_sigmas
        return standardize.decode(to_diffused_xs)

        # to_diffused_xs = self.diffused_xs + self.eps * (to_sigmas - self.from_sigmas)
        # return standardize.decode(to_diffused_xs)

    def correction(self, previous_diffused_images, previous_ts, previous_eps):
        previous_diffused_xs = standardize.encode(
            previous_diffused_images.to(self.device)
        )
        corrected_diffused_xs = (
            previous_diffused_xs
            + (self.from_sigmas - self.sigmas(previous_ts))
            * (self.eps + previous_eps)
            / 2
        )
        return standardize.decode(corrected_diffused_xs)



from .adabins_depth import AdaBinsDepth
from .blip import BLIP
from .clip import CLIP
from .cloob import CLOOB
from .deep_image_prior import DeepImagePrior
from .glide_clip import GlideCLIP
from .guided_diffusion import GuidedDiffusion
from . import latent_diffusion
from .midas_depth import MidasDepth
from .monster_diffusion import MonsterDiffusion
from .owlvit import OWLViT
from .ruclip import RuCLIP
from .slip import SLIP
from .stable_diffusion import StableDiffusion
from .stylegan_xl import StyleGANXL
from .super_resolution import SuperResolution
from .velocity_diffusion import VelocityDiffusion
from .open_clip import OpenCLIP
from .simulacra_aesthetic import SimulacraAesthetic



from contextlib import contextmanager
import torch
import torch.nn as nn
import torch.nn.functional as F

from .interface import LossInterface
from perceptor import models, transforms
from perceptor.models.velocity_diffusion import diffusion_space


class VelocityDiffusion(LossInterface):
    def __init__(self, model, noise, from_ts=0.5, resample_ts=0.3):
        super().__init__()
        self.from_ts = from_ts
        self.resample_ts = resample_ts
        self.model = model
        self.noise = nn.Parameter(noise, requires_grad=True)

    def diffuse_denoise(self, denoised, **extra_kwargs):
        predictions = self.model.predictions(
            self.model.diffuse(denoised, self.from_ts, noise=self.noise),
            self.from_ts,
            **extra_kwargs,
        )
        return predictions.denoised_images

    def forward(self, images, frozen_diffused_denoised):
        return F.mse_loss(
            frozen_diffused_denoised.detach().clamp(0, 1),
            transforms.clamp_with_grad(images),
        )

    @contextmanager
    def guided_resample_(
        self, denoised, guidance_scale=0.5, clamp_value=1e-6, **extra_kwargs
    ):
        """
        Resamples noise in direction of the gradient

        Usage:

            with diffusion.guided_resample_(images) as diffused_denoised:
                clip(diffused_denoised).backward()
        """
        if self.noise.grad is not None:
            self.noise.grad.zero_()
        with torch.enable_grad():
            from_diffused = self.model.diffuse(denoised, self.from_ts, noise=self.noise)
            predictions = self.model.predictions(
                from_diffused, self.from_ts, **extra_kwargs
            )
            diffuse_denoise = predictions.denoised_images
            yield diffuse_denoise
        guided_predictions = predictions.guided(
            -self.noise.grad, guidance_scale=guidance_scale, clamp_value=clamp_value
        )
        self.noise.data = guided_predictions.resample_noise(
            self.resample_ts,
            **extra_kwargs,
        )
        self.noise.grad.zero_()

    def compensate_noise_(self, from_denoised, to_denoised):
        old_pred = diffusion_space.encode(from_denoised)
        new_pred = diffusion_space.encode(to_denoised)
        delta_pred = new_pred - old_pred
        self.noise.data = self.noise.data - delta_pred.data

    def noise_step_(self, from_denoised, from_t, to_t, to_denoised):
        """
        1. Step to to_t
        2. Add noise to from_t
        """
        from_diffused = self.model.diffuse(from_denoised, from_t, noise=self.noise)
        forced_from_eps = self.model.forced_eps(from_diffused, from_t, to_denoised)
        # should be same as self.noise?
        assert forced_from_eps == self.noise
        to_diffused = self.model.step(from_diffused, from_t, to_t, to_denoised)
        reversed_diffused = self.model.reverse_step(to_diffused, to_t, from_t)
        reversed_eps = self.model.eps(reversed_diffused, from_t)
        self.noise.data = reversed_eps


def test_velocity_diffusion_loss():
    model = models.VelocityDiffusion().cuda()
    diffusion_loss = VelocityDiffusion(model, torch.randn((1, *model.shape))).cuda()
    images = torch.zeros((1, *model.shape)).cuda()

    with diffusion_loss.guided_resample_(images) as diffused_denoised:
        diffused_denoised.square().mean().backward()



import numpy as np
from torch import nn
from basicsr.utils.download_util import load_file_from_url
from lantern import Tensor

from perceptor import utils, transforms
from .infer import InferenceHelper

MAX_ADABINS_AREA = 500000
MIN_ADABINS_AREA = 448 * 448

checkpoints = dict(
    nyu="https://s3.eu-central-1.wasabisys.com/nextml-model-data/ada-bins/AdaBins_nyu.pt",
    kitti="https://s3.eu-central-1.wasabisys.com/nextml-model-data/ada-bins/AdaBins_kitti.pt",
)


@utils.cache
class AdaBinsDepth(nn.Module):
    def __init__(self, name="nyu"):
        """
        Args:
            name (str, optional): Name of the model. Available weights: "nyu" and "kitti"
        """
        super().__init__()
        self.name = name
        checkpoint_path = load_file_from_url(
            checkpoints[self.name],
            "models",
        )
        self.model = InferenceHelper(name, checkpoint_path)
        self.model.eval()
        self.model.requires_grad_(False)

    def forward(self, images: Tensor.dims("NCHW")) -> Tensor.dims("NCHW"):
        h, w = images.shape[-2:]
        image_area = w * h
        if image_area > MAX_ADABINS_AREA:
            scale = np.sqrt(MAX_ADABINS_AREA) / np.sqrt(image_area)
            images = transforms.resize(
                images, out_shape=(int(w * scale), int(h * scale)), resample="lancsoz3"
            )
        elif image_area < MIN_ADABINS_AREA:
            scale = np.sqrt(MIN_ADABINS_AREA) / np.sqrt(image_area)
            images = transforms.resize(
                images, out_shape=(int(w * scale), int(h * scale)), resample="bicubic"
            )

        return self.model.predict(images)


def test_adabins_depth():
    import requests
    from PIL import Image
    import torch
    import torchvision.transforms.functional as TF

    image_url = "http://images.cocodataset.org/val2017/000000039769.jpg"
    images = (
        TF.to_tensor(Image.open(requests.get(image_url, stream=True).raw))[None]
        .cuda()
        .requires_grad_()
    )

    model = AdaBinsDepth().cuda()

    with torch.enable_grad():
        depths = model(images)
        depths.mean().backward()

    assert images.grad is not None

    utils.pil_image(
        torch.cat(
            [
                images,
                (depths.repeat(1, 3, 1, 1) - depths.min())
                / (depths.max() - depths.min()),
            ]
        )
    ).save("tests/adabins_depth.png")



from readline import set_completer_delims
import sys
from pathlib import Path
import dill
import numpy as np
import torch
import torch.nn as nn
from basicsr.utils.download_util import load_file_from_url

from perceptor import utils
from perceptor.models.stylegan_xl import (
    gen_utils,
    torch_utils,
    dnnlib,
    legacy,
)

checkpoints = dict(
    imagenet128="https://s3.eu-central-1.wasabisys.com/nextml-model-data/stylegan-xl/imagenet128.dill",
    ffhq256="https://s3.eu-central-1.wasabisys.com/nextml-model-data/stylegan-xl/ffhq256.dill",
    pokemon256="https://s3.eu-central-1.wasabisys.com/nextml-model-data/stylegan-xl/pokemon256.dill",
)


@utils.cache
class StyleGANXL(nn.Module):
    def __init__(self, name="imagenet128"):
        """
        Args:
            name (str, optional): Name of the model. imagenet128, ffhq256, pokemon256
        """
        super().__init__()
        self.name = name

        checkpoint_path = load_file_from_url(
            checkpoints[self.name],
            "models",
        )

        # hack for dill/pickle
        sys.modules["torch_utils"] = torch_utils
        sys.modules["dnnlib"] = dnnlib
        sys.modules["legacy"] = legacy

        self.model = dill.load(Path(checkpoint_path).open("rb"))
        self.model.eval().requires_grad_(False)

        # TODO: half precision is supported

    @property
    def device(self):
        return next(iter(self.parameters()))

    def forward(self, latents):
        return gen_utils.w_to_img(self.model, latents).add(1).div(2)

    def latents(self, size, seeds=None, class_indices=None):
        if class_indices is None:
            class_indices = [None for _ in range(size)]

        return torch.cat(
            [
                gen_utils.get_w_from_seed(
                    self.model,
                    1,
                    self.device,
                    truncation_psi=1,
                    seed=(None if seeds is None else seeds[i]),
                    class_idx=(None if class_indices is None else class_indices[i]),
                )
                for i in range(size)
            ]
        )



# -*- coding: utf-8 -*-
import math
import torch
from torch import nn as nn
from torch.nn import functional as F
from torch.nn import init as init
from torch.nn.modules.batchnorm import _BatchNorm


@torch.no_grad()
def default_init_weights(module_list, scale=1, bias_fill=0, **kwargs):
    """Initialize network weights.
    Args:
        module_list (list[nn.Module] | nn.Module): Modules to be initialized.
        scale (float): Scale initialized weights, especially for residual
            blocks. Default: 1.
        bias_fill (float): The value to fill bias. Default: 0
        kwargs (dict): Other arguments for initialization function.
    """
    if not isinstance(module_list, list):
        module_list = [module_list]
    for module in module_list:
        for m in module.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight, **kwargs)
                m.weight.data *= scale
                if m.bias is not None:
                    m.bias.data.fill_(bias_fill)
            elif isinstance(m, nn.Linear):
                init.kaiming_normal_(m.weight, **kwargs)
                m.weight.data *= scale
                if m.bias is not None:
                    m.bias.data.fill_(bias_fill)
            elif isinstance(m, _BatchNorm):
                init.constant_(m.weight, 1)
                if m.bias is not None:
                    m.bias.data.fill_(bias_fill)


def make_layer(basic_block, num_basic_block, **kwarg):
    """Make layers by stacking the same blocks.
    Args:
        basic_block (nn.module): nn.module class for basic block.
        num_basic_block (int): number of blocks.
    Returns:
        nn.Sequential: Stacked blocks in nn.Sequential.
    """
    layers = []
    for _ in range(num_basic_block):
        layers.append(basic_block(**kwarg))
    return nn.Sequential(*layers)


class ResidualBlockNoBN(nn.Module):
    """Residual block without BN.
    It has a style of:
        ---Conv-ReLU-Conv-+-
         |________________|
    Args:
        num_feat (int): Channel number of intermediate features.
            Default: 64.
        res_scale (float): Residual scale. Default: 1.
        pytorch_init (bool): If set to True, use pytorch default init,
            otherwise, use default_init_weights. Default: False.
    """

    def __init__(self, num_feat=64, res_scale=1, pytorch_init=False):
        super(ResidualBlockNoBN, self).__init__()
        self.res_scale = res_scale
        self.conv1 = nn.Conv2d(num_feat, num_feat, 3, 1, 1, bias=True)
        self.conv2 = nn.Conv2d(num_feat, num_feat, 3, 1, 1, bias=True)
        self.relu = nn.ReLU(inplace=True)

        if not pytorch_init:
            default_init_weights([self.conv1, self.conv2], 0.1)

    def forward(self, x):
        identity = x
        out = self.conv2(self.relu(self.conv1(x)))
        return identity + out * self.res_scale


class Upsample(nn.Sequential):
    """Upsample module.
    Args:
        scale (int): Scale factor. Supported scales: 2^n and 3.
        num_feat (int): Channel number of intermediate features.
    """

    def __init__(self, scale, num_feat):
        m = []
        if (scale & (scale - 1)) == 0:  # scale = 2^n
            for _ in range(int(math.log(scale, 2))):
                m.append(nn.Conv2d(num_feat, 4 * num_feat, 3, 1, 1))
                m.append(nn.PixelShuffle(2))
        elif scale == 3:
            m.append(nn.Conv2d(num_feat, 9 * num_feat, 3, 1, 1))
            m.append(nn.PixelShuffle(3))
        else:
            raise ValueError(
                f"scale {scale} is not supported. " "Supported scales: 2^n and 3."
            )
        super(Upsample, self).__init__(*m)


def flow_warp(
    x, flow, interp_mode="bilinear", padding_mode="zeros", align_corners=True
):
    """Warp an image or feature map with optical flow.
    Args:
        x (Tensor): Tensor with size (n, c, h, w).
        flow (Tensor): Tensor with size (n, h, w, 2), normal value.
        interp_mode (str): 'nearest' or 'bilinear'. Default: 'bilinear'.
        padding_mode (str): 'zeros' or 'border' or 'reflection'.
            Default: 'zeros'.
        align_corners (bool): Before pytorch 1.3, the default value is
            align_corners=True. After pytorch 1.3, the default value is
            align_corners=False. Here, we use the True as default.
    Returns:
        Tensor: Warped image or feature map.
    """
    assert x.size()[-2:] == flow.size()[1:3]
    _, _, h, w = x.size()
    # create mesh grid
    grid_y, grid_x = torch.meshgrid(
        torch.arange(0, h).type_as(x), torch.arange(0, w).type_as(x)
    )
    grid = torch.stack((grid_x, grid_y), 2).float()  # W(x), H(y), 2
    grid.requires_grad = False

    vgrid = grid + flow
    # scale grid to [-1,1]
    vgrid_x = 2.0 * vgrid[:, :, :, 0] / max(w - 1, 1) - 1.0
    vgrid_y = 2.0 * vgrid[:, :, :, 1] / max(h - 1, 1) - 1.0
    vgrid_scaled = torch.stack((vgrid_x, vgrid_y), dim=3)
    output = F.grid_sample(
        x,
        vgrid_scaled,
        mode=interp_mode,
        padding_mode=padding_mode,
        align_corners=align_corners,
    )

    # TODO, what if align_corners=False
    return output


def resize_flow(flow, size_type, sizes, interp_mode="bilinear", align_corners=False):
    """Resize a flow according to ratio or shape.
    Args:
        flow (Tensor): Precomputed flow. shape [N, 2, H, W].
        size_type (str): 'ratio' or 'shape'.
        sizes (list[int | float]): the ratio for resizing or the final output
            shape.
            1) The order of ratio should be [ratio_h, ratio_w]. For
            downsampling, the ratio should be smaller than 1.0 (i.e., ratio
            < 1.0). For upsampling, the ratio should be larger than 1.0 (i.e.,
            ratio > 1.0).
            2) The order of output_size should be [out_h, out_w].
        interp_mode (str): The mode of interpolation for resizing.
            Default: 'bilinear'.
        align_corners (bool): Whether align corners. Default: False.
    Returns:
        Tensor: Resized flow.
    """
    _, _, flow_h, flow_w = flow.size()
    if size_type == "ratio":
        output_h, output_w = int(flow_h * sizes[0]), int(flow_w * sizes[1])
    elif size_type == "shape":
        output_h, output_w = sizes[0], sizes[1]
    else:
        raise ValueError(
            f"Size type should be ratio or shape, but got type {size_type}."
        )

    input_flow = flow.clone()
    ratio_h = output_h / flow_h
    ratio_w = output_w / flow_w
    input_flow[:, 0, :, :] *= ratio_w
    input_flow[:, 1, :, :] *= ratio_h
    resized_flow = F.interpolate(
        input=input_flow,
        size=(output_h, output_w),
        mode=interp_mode,
        align_corners=align_corners,
    )
    return resized_flow


# TODO: may write a cpp file
def pixel_unshuffle(x, scale):
    """Pixel unshuffle.
    Args:
        x (Tensor): Input feature with shape (b, c, hh, hw).
        scale (int): Downsample ratio.
    Returns:
        Tensor: the pixel unshuffled feature.
    """
    b, c, hh, hw = x.size()
    out_channel = c * (scale**2)
    assert hh % scale == 0 and hw % scale == 0
    h = hh // scale
    w = hw // scale
    x_view = x.view(b, c, h, scale, w, scale)
    return x_view.permute(0, 1, 3, 5, 2, 4).reshape(b, out_channel, h, w)



from functools import partial
import torch
import lantern
from basicsr.utils.download_util import load_file_from_url

from perceptor.utils import cache
from perceptor import models
from .models import get_model
from .model_urls import model_urls
from . import diffusion_space, utils
from .predictions import Predictions


class Model(torch.nn.Module):
    def __init__(self, name="yfcc_2"):
        """
        Args:
            name: The name of the model. Available models are:
                - yfcc_2
                - yfcc_1
                - cc12m_1_cfg (conditioned)
                - wikiart
        """
        super().__init__()
        self.name = name
        self.model = get_model(name)()
        checkpoint_path = load_file_from_url(model_urls[name], "models")
        self.model.load_state_dict(torch.load(checkpoint_path, map_location="cpu"))
        print(f"Loaded checkpoint {checkpoint_path}")
        self.model.eval()
        self.model.requires_grad_(False)

    def to(self, device):
        super().to(device)
        if device == torch.device("cuda"):
            self.model.half()
        return self

    @property
    def device(self):
        return next(iter(self.parameters())).device

    @property
    def shape(self):
        return self.model.shape

    @staticmethod
    def schedule_ts(n_steps=500, from_ts=1.0, to_ts=1e-2, rho=7.0):
        from_alpha, from_sigma = utils.t_to_alpha_sigma(torch.as_tensor(from_ts))
        to_alpha, to_sigma = utils.t_to_alpha_sigma(torch.as_tensor(to_ts))

        from_log_snr = utils.alpha_sigma_to_log_snr(from_alpha, from_sigma)
        to_log_snr = utils.alpha_sigma_to_log_snr(to_alpha, to_sigma)

        elucidated_from_sigma = (1 / from_log_snr.exp()).sqrt().clamp(max=150)
        elucidated_to_sigma = (1 / to_log_snr.exp()).sqrt().clamp(min=1e-3)

        ramp = torch.linspace(0, 1, n_steps + 1)
        min_inv_rho = elucidated_to_sigma ** (1 / rho)
        max_inv_rho = elucidated_from_sigma ** (1 / rho)
        sigmas = (max_inv_rho + ramp * (min_inv_rho - max_inv_rho)) ** rho
        log_snr = utils.alpha_sigma_to_log_snr(torch.ones_like(sigmas), sigmas)
        alpha, sigma = utils.log_snr_to_alpha_sigma(log_snr)
        schedule_ts = utils.alpha_sigma_to_t(alpha, sigma)
        return torch.stack([schedule_ts[:-1], schedule_ts[1:]], dim=1)

    def random_diffused(self, shape):
        return diffusion_space.decode(torch.randn(shape)).to(self.device)

    @staticmethod
    def sigmas_to_ts(sigmas):
        sigmas = torch.as_tensor(sigmas)
        return utils.sigma_to_t(sigmas)

    def alphas(self, ts):
        if isinstance(ts, float):
            ts = torch.tensor(ts)
        if ts.ndim == 0:
            ts = ts[None]
        if ts.ndim != 1:
            raise ValueError("t must be a scalar or a 1D tensor")
        alphas, _ = utils.t_to_alpha_sigma(ts)
        return alphas[:, None, None, None].to(self.device)

    def sigmas(self, ts):
        if isinstance(ts, float):
            ts = torch.tensor(ts)
        if ts.ndim == 0:
            ts = ts[None]
        if ts.ndim != 1:
            raise ValueError("t must be a scalar or a 1D tensor")
        _, sigmas = utils.t_to_alpha_sigma(ts)
        return sigmas[:, None, None, None].to(self.device)

    @torch.cuda.amp.autocast()
    def velocities(self, diffused, t, conditioning=None):
        x = diffusion_space.encode(diffused)

        if hasattr(self.model, "clip_model"):
            model_fn = partial(self.model, clip_embed=conditioning.squeeze(dim=1))
        else:
            model_fn = self.model

        if isinstance(t, float) or t.ndim == 0:
            t = torch.full((x.shape[0],), t).to(x)

        velocities = model_fn(x, t)
        return velocities.float()

    def forward(self, diffused_images, ts, conditioning=None):
        if isinstance(ts, float) or ts.ndim == 0:
            ts = torch.full((diffused_images.shape[0],), ts).to(diffused_images)
        return Predictions(
            from_diffused_images=diffused_images,
            from_ts=ts,
            velocities=self.velocities(diffused_images, ts, conditioning),
        )

    def predictions(self, diffused_images, ts, conditioning=None):
        return self.forward(diffused_images, ts, conditioning)

    def conditioning(self, texts=None, images=None, encodings=None):
        clip_model = models.CLIP(self.model.clip_model)

        all_encodings = list()
        if texts is not None:
            all_encodings.append(clip_model.encode_texts(texts))
        if images is not None:
            all_encodings.append(clip_model.encode_images(images))
        if encodings is not None:
            all_encodings.append(encodings)
        if len(all_encodings) == 0:
            raise ValueError("Must provide at least one of texts, images, or encodings")
        return torch.stack(all_encodings, dim=0).mean(dim=0)[None]

    def diffuse(self, denoised_images, ts, noise=None):
        denoised_xs = diffusion_space.encode(denoised_images)
        if isinstance(ts, float) or ts.ndim == 0:
            ts = torch.full((denoised_xs.shape[0],), ts).to(denoised_xs)
        if noise is None:
            noise = torch.randn_like(denoised_xs)
        alphas, sigmas = self.alphas(ts), self.sigmas(ts)
        return diffusion_space.decode(denoised_xs * alphas + noise * sigmas)

    def inject_noise(
        self, diffused_images, ts, reversed_ts, extra_noise_multiplier=1.003
    ):
        diffused_xs = diffusion_space.encode(diffused_images).to(self.device)

        diffused_multiplier = self.alphas(reversed_ts) / self.alphas(ts)
        target_sigmas = self.sigmas(reversed_ts)
        additional_noise_std = (
            target_sigmas.square()
            - self.sigmas(ts).square() * diffused_multiplier.square()
        ).sqrt()

        reversed_diffused_xs = (
            diffused_xs * diffused_multiplier
            + additional_noise_std
            * torch.randn_like(diffused_xs)
            * extra_noise_multiplier
        )
        return diffusion_space.decode(reversed_diffused_xs)


VelocityDiffusion: Model = cache(Model)


def test_velocity_diffusion():
    from perceptor import utils

    torch.set_grad_enabled(False)
    device = torch.device("cuda")

    diffusion = models.VelocityDiffusion("yfcc_2").to(device)

    diffused_images = diffusion.random_diffused((1, 3, 512, 512)).to(device)

    for from_ts, to_ts in diffusion.schedule_ts(n_steps=50):
        if (from_ts < 1.0).all():
            new_from_ts = from_ts * 1.003
            diffused_images = diffusion.predictions(
                diffused_images, from_ts
            ).noisy_reverse_step(new_from_ts)
            from_ts = new_from_ts

        predictions = diffusion.predictions(
            diffused_images,
            from_ts,
        )
        diffused_images = predictions.step(to_ts)
        diffused_images = (
            diffusion.predictions(diffused_images, to_ts)
            .correction(predictions)
            .step(to_ts)
        )

    utils.pil_image(diffusion.predictions(diffused_images, to_ts).denoised_images).save(
        "tests/velocity_diffusion_yfcc_2.png"
    )


def test_conditioned_velocity_diffusion():
    from perceptor import utils

    torch.set_grad_enabled(False)
    device = torch.device("cuda")

    diffusion = VelocityDiffusion("cc12m_1_cfg").to(device)

    diffused_images = diffusion.random_diffused((1, 3, 256, 256)).to(device)

    conditioning = diffusion.conditioning(texts=["photo of a cute cat"])

    for from_ts, to_ts in diffusion.schedule_ts(n_steps=50):
        predictions = diffusion.predictions(diffused_images, from_ts, conditioning)
        diffused_images = predictions.step(to_ts)

    utils.pil_image(
        diffusion.predictions(diffused_images, to_ts, conditioning).denoised_images
    ).save("tests/velocity_diffusion_cc12m_1.png")


def test_convert_sigma_ts():
    diffusion = VelocityDiffusion("cc12m_1_cfg")
    from_ts = 0.3
    assert from_ts == diffusion.sigmas_to_ts(diffusion.sigmas(from_ts))


def test_schedule_ts():
    diffusion = VelocityDiffusion("cc12m_1_cfg")
    from_ts = 0.6
    assert torch.allclose(
        diffusion.schedule_ts(n_steps=50, from_ts=from_ts)[0, 0],
        torch.as_tensor(from_ts),
    )


def test_utils_conversion():
    t = torch.as_tensor(0.3)
    alpha, sigma = utils.t_to_alpha_sigma(t)
    assert torch.allclose(utils.sigma_to_t(sigma), t)
    assert t == utils.alpha_sigma_to_t(alpha, sigma)



from lantern import FunctionalBase, Tensor
import torch


class GradientCheckpoint(FunctionalBase):
    original: Tensor
    detached: Tensor

    def __init__(self, tensor):
        super().__init__(original=tensor, detached=tensor.detach().requires_grad_())

    def zero_grad_(self):
        self.detached.grad.zero_()
        return self

    def backward(self, loss):
        loss.backward()
        gradients = self.detached.grad.clone()
        self.zero_grad_()
        return gradients

    def continue_backward(self, gradients=None, retain_graph=False):
        if self.detached.grad is None:
            raise ValueError("Gradient is not defined")

        if gradients is None:
            return self.original.backward(self.detached.grad, retain_graph=retain_graph)
        else:
            return self.original.backward(gradients, retain_graph=retain_graph)

    def tensor(self):
        return self.detached

    @staticmethod
    def nonzero_mean(gradients, dim=0):
        if isinstance(gradients, list):
            gradients = torch.stack(gradients)
        return gradients.sum(dim).div(gradients.ne(0).sum(dim).add(1e-6))

    @staticmethod
    def nonzero_scale(tensor, dim=None):
        if isinstance(tensor, list):
            tensor = torch.stack(tensor)
        shape = tensor.shape
        if dim is None:
            tensor = tensor.flatten()
            dim = 0

        mask = tensor.ne(0)
        mean_square = tensor.square().sum(dim) / mask.sum(dim).add(1e-6)
        mean = tensor.sum(dim) / mask.sum(dim).add(1e-6)
        std = (mean_square - mean.square()).sqrt().add(1e-6)
        scaled_tensor = tensor / std.unsqueeze(dim).add(1e-6)
        return scaled_tensor.view(*shape)


def gradient_checkpoint(tensor: Tensor) -> GradientCheckpoint:
    """
    Gradient checkpointing to save compute for common part of graph.

    Usage:

        checkpoint = gradient_checkpoint(images)
        for text_loss in text_losses:
            text_loss(checkpoint.tensor()).backward()
        checkpoint.continue_backward()
    """
    return GradientCheckpoint(tensor)


def test_gradient_checkpoint():
    import torch

    with torch.enable_grad():
        images = torch.zeros(1, 3, 64, 64).requires_grad_()
        checkpoint = gradient_checkpoint(images * 2)
        checkpoint.tensor().pow(2).mean().backward()
        assert checkpoint.detached.grad is not None
        checkpoint.continue_backward()
    assert images.grad is not None



import torch
import torch.nn as nn


class PatchTransformerEncoder(nn.Module):
    def __init__(self, in_channels, patch_size=10, embedding_dim=128, num_heads=4):
        super(PatchTransformerEncoder, self).__init__()
        encoder_layers = nn.TransformerEncoderLayer(
            embedding_dim, num_heads, dim_feedforward=1024
        )
        self.transformer_encoder = nn.TransformerEncoder(
            encoder_layers, num_layers=4
        )  # takes shape S,N,E

        self.embedding_convPxP = nn.Conv2d(
            in_channels,
            embedding_dim,
            kernel_size=patch_size,
            stride=patch_size,
            padding=0,
        )

        self.positional_encodings = nn.Parameter(
            torch.rand(500, embedding_dim), requires_grad=True
        )

    def forward(self, x):
        embeddings = self.embedding_convPxP(x).flatten(
            2
        )  # .shape = n,c,s = n, embedding_dim, s
        # embeddings = nn.functional.pad(embeddings, (1,0))  # extra special token at start ?
        embeddings = embeddings + self.positional_encodings[
            : embeddings.shape[2], :
        ].T.unsqueeze(0)

        # change to S,N,E format required by transformer
        embeddings = embeddings.permute(2, 0, 1)
        x = self.transformer_encoder(embeddings)  # .shape = S, N, E
        return x


class PixelWiseDotProduct(nn.Module):
    def __init__(self):
        super(PixelWiseDotProduct, self).__init__()

    def forward(self, x, K):
        n, c, h, w = x.size()
        _, cout, ck = K.size()
        assert (
            c == ck
        ), "Number of channels in x and Embedding dimension (at dim 2) of K matrix must match"
        y = torch.matmul(
            x.view(n, c, h * w).permute(0, 2, 1), K.permute(0, 2, 1)
        )  # .shape = n, hw, cout
        return y.permute(0, 2, 1).view(n, cout, h, w)



import importlib

import torch
import numpy as np

from inspect import isfunction
from PIL import Image, ImageDraw, ImageFont


def log_txt_as_img(wh, xc, size=10):
    # wh a tuple of (width, height)
    # xc a list of captions to plot
    b = len(xc)
    txts = list()
    for bi in range(b):
        txt = Image.new("RGB", wh, color="white")
        draw = ImageDraw.Draw(txt)
        font = ImageFont.truetype("data/DejaVuSans.ttf", size=size)
        nc = int(40 * (wh[0] / 256))
        lines = "\n".join(
            xc[bi][start : start + nc] for start in range(0, len(xc[bi]), nc)
        )

        try:
            draw.text((0, 0), lines, fill="black", font=font)
        except UnicodeEncodeError:
            print("Cant encode string for logging. Skipping.")

        txt = np.array(txt).transpose(2, 0, 1) / 127.5 - 1.0
        txts.append(txt)
    txts = np.stack(txts)
    txts = torch.tensor(txts)
    return txts


def ismap(x):
    if not isinstance(x, torch.Tensor):
        return False
    return (len(x.shape) == 4) and (x.shape[1] > 3)


def isimage(x):
    if not isinstance(x, torch.Tensor):
        return False
    return (len(x.shape) == 4) and (x.shape[1] == 3 or x.shape[1] == 1)


def exists(x):
    return x is not None


def default(val, d):
    if exists(val):
        return val
    return d() if isfunction(d) else d


def mean_flat(tensor):
    """
    https://github.com/openai/guided-diffusion/blob/27c20a8fab9cb472df5d6bdd6c8d11c8f430b924/guided_diffusion/nn.py#L86
    Take the mean over all non-batch dimensions.
    """
    return tensor.mean(dim=list(range(1, len(tensor.shape))))


def count_params(model, verbose=False):
    total_params = sum(p.numel() for p in model.parameters())
    if verbose:
        print(f"{model.__class__.__name__} has {total_params*1.e-6:.2f} M params.")
    return total_params


def instantiate_from_config(config):
    if not "target" in config:
        if config == "__is_first_stage__":
            return None
        elif config == "__is_unconditional__":
            return None
        raise KeyError("Expected key `target` to instantiate.")
    if config["target"][: len("ldm.")] == "ldm.":
        target = f"perceptor.models.latent_diffusion.{config['target']}"
    else:
        target = config["target"]
    return get_obj_from_str(target)(**config.get("params", dict()))


def get_obj_from_str(string, reload=False):
    module, cls = string.rsplit(".", 1)
    if reload:
        module_imp = importlib.import_module(module)
        importlib.reload(module_imp)
    return getattr(importlib.import_module(module, package=None), cls)



import numpy as np
import torch


def gradient_2d(start, stop, width, height, is_horizontal):
    if is_horizontal:
        return np.tile(np.linspace(start, stop, width), (height, 1))
    else:
        return np.tile(np.linspace(start, stop, height), (width, 1)).T


def gradient_3d(width, height, start_list, stop_list, is_horizontal_list):
    result = np.zeros((height, width, len(start_list)), dtype=float)

    for i, (start, stop, is_horizontal) in enumerate(
        zip(start_list, stop_list, is_horizontal_list)
    ):
        result[:, :, i] = gradient_2d(start, stop, width, height, is_horizontal)

    return result


def gradient(shape):
    n, c, h, w = shape
    if c != 3:
        raise ValueError("Only 3 channel images are supported.")

    return (
        torch.from_numpy(
            np.stack(
                [
                    gradient_3d(
                        w,
                        h,
                        (0, 0, np.random.randint(0, 255)),
                        (
                            np.random.randint(1, 255),
                            np.random.randint(2, 255),
                            np.random.randint(3, 128),
                        ),
                        (True, False, False),
                    )
                    / 255
                    for _ in range(n)
                ]
            )
        )
        .float()
        .permute(0, 3, 1, 2)
    )



# https://github.com/assafshocher/ResizeRight/blob/master/interp_methods.py
from configparser import Interpolation
from math import pi

try:
    import torch
except ImportError:
    torch = None

try:
    import numpy
except ImportError:
    numpy = None

if numpy is None and torch is None:
    raise ImportError("Must have either Numpy or PyTorch but both not found")


def set_framework_dependencies(x):
    if type(x) is numpy.ndarray:
        to_dtype = lambda a: a
        fw = numpy
    else:
        to_dtype = lambda a: a.to(x.dtype)
        fw = torch
    eps = fw.finfo(fw.float32).eps
    return fw, to_dtype, eps


def support_sz(sz):
    def wrapper(f):
        f.support_sz = sz
        return f

    return wrapper


@support_sz(4)
def cubic(x):
    fw, to_dtype, eps = set_framework_dependencies(x)
    absx = fw.abs(x)
    absx2 = absx**2
    absx3 = absx**3
    return (1.5 * absx3 - 2.5 * absx2 + 1.0) * to_dtype(absx <= 1.0) + (
        -0.5 * absx3 + 2.5 * absx2 - 4.0 * absx + 2.0
    ) * to_dtype((1.0 < absx) & (absx <= 2.0))


@support_sz(4)
def lanczos2(x):
    fw, to_dtype, eps = set_framework_dependencies(x)
    return (
        (fw.sin(pi * x) * fw.sin(pi * x / 2) + eps) / ((pi**2 * x**2 / 2) + eps)
    ) * to_dtype(abs(x) < 2)


@support_sz(6)
def lanczos3(x):
    fw, to_dtype, eps = set_framework_dependencies(x)
    return (
        (fw.sin(pi * x) * fw.sin(pi * x / 3) + eps) / ((pi**2 * x**2 / 3) + eps)
    ) * to_dtype(abs(x) < 3)


@support_sz(2)
def linear(x):
    fw, to_dtype, eps = set_framework_dependencies(x)
    return (x + 1) * to_dtype((-1 <= x) & (x < 0)) + (1 - x) * to_dtype(
        (0 <= x) & (x <= 1)
    )


@support_sz(1)
def box(x):
    fw, to_dtype, eps = set_framework_dependencies(x)
    return to_dtype((-1 <= x) & (x < 0)) + to_dtype((0 <= x) & (x <= 1))


methods = dict(
    cubic=cubic,
    bicubic=cubic,
    linear=linear,
    bilinear=linear,
    lanczos2=lanczos2,
    lanczos3=lanczos3,
    box=box,
)



P_mean = -1.2
P_std = 1.2
sigma_data = 0.5
rho = 7
sigma_min = 1e-2  # 0.002
sigma_max = 80

# depends on dataset, see table 5 in https://arxiv.org/pdf/2206.00364.pdf
# ImageNet 64x64
S_tmin = 0.05
S_tmax = 50
S_churn = 80
S_noise = 1.003  # between 1.0 and 1.007



from torch import nn


class DrawingInterface(nn.Module):
    def forward(self, *args, **kwargs):
        return self.synthesize(*args, **kwargs)

    def synthesize(self, *args, **kwargs):
        raise NotImplementedError

    def encode(self, *args, **kwargs):
        raise NotImplementedError

    def replace_(self, *args, **kwargs):
        raise NotImplementedError



import torch.nn as nn
from torchvision import models


class VGG19(nn.Module):
    def __init__(self, name="squeeze", linear_layers=True, spatial=False):
        super().__init__()
        self.model = models.vgg19(pretrained=True).features
        self.model.eval()
        self.model.requires_grad_(False)

    def forward(self, images):
        """
        Args:
            images_a: images of shape (batch_size, 3, height, width) between 0 and 1
            images_b: images of shape (batch_size, 3, height, width) between 0 and 1
        """

        _, _, height, width = images.shape
        if width % 8 != 0:
            raise ValueError("Width must be divisible by 8")
        if height % 8 != 0:
            raise ValueError("Height must be divisible by 8")

        return self.model(images)


def test_vgg19():
    import torch

    model = VGG19()
    images_a = torch.randn((1, 3, 256, 256))
    model(images_a)



import torch
from torch import nn
from basicsr.utils.download_util import load_file_from_url

from perceptor.utils import cache
from . import diffusion_space
from .unet import UNetModel
from .script_util import (
    create_model_and_diffusion,
    model_and_diffusion_defaults,
)


@cache
class GuidedDiffusion(nn.Module):
    def __init__(self, name="standard"):
        """
        Args:
            name: The name of the model.

        Usage:

            diffusion = models.GuidedDiffusion("pixelart").to(device)

            diffused_image = torch.randn((1, 3, 256, 256))

            for from_index, to_index in model.schedule_indices():
                eps = diffusion.eps(diffused_image, from_index)
                denoised_image = diffusion.denoise(diffused_image, from_index, eps)
                diffused_image = diffusion.step(diffused_image, eps, from_index, to_index)
            denoised_image = diffusion.denoise(diffused_image, to_index)
        """
        super().__init__()
        self.name = name

        if name == "standard":
            self.model, self.diffusion = create_openimages_model()
            checkpoint_path = load_file_from_url(
                "https://huggingface.co/lowlevelware/512x512_diffusion_unconditional_ImageNet/resolve/main/512x512_diffusion_uncond_finetune_008100.pt",
                # alternative: "https://set.zlkj.in/models/diffusion/512x512_diffusion_uncond_openimages_epoch28_withfilter.pt",
                "models",
            )
            self.shape = (3, 512, 512)
        elif name == "pixelart":
            self.model, self.diffusion = create_pixelart_model()
            checkpoint_path = load_file_from_url(
                "https://huggingface.co/KaliYuga/PADexpanded/resolve/main/PADexpanded.pt",
                "models",
            )
            self.shape = (3, 256, 256)
        else:
            raise ValueError(f"Unknown model name {self.name}")

        self.model.load_state_dict(torch.load(checkpoint_path, map_location="cpu"))
        self.model.requires_grad_(False).eval()

    @property
    def device(self):
        return next(iter(self.parameters())).device

    def schedule_indices(self, from_index=999, to_index=20, n_steps=None):
        if from_index < to_index:
            raise ValueError("from_index must be greater than to_index")
        if n_steps is None:
            n_steps = (from_index - to_index) // 2
        schedule_indices = torch.linspace(from_index, to_index, n_steps).long()
        from_indices = schedule_indices[:-1]
        to_indices = schedule_indices[1:]
        if (from_indices == to_indices).any():
            raise ValueError("Schedule indices must be unique")
        return torch.stack([from_indices, to_indices], dim=1)

    def random_diffused(self, shape):
        return diffusion_space.decode(torch.randn(shape)).to(self.device)

    def forward(self, diffused, from_index):
        return self.denoise(diffused, from_index)

    def denoise(self, diffused, from_index, eps=None):
        x = diffusion_space.encode(diffused)
        if isinstance(from_index, int) or from_index.ndim == 0:
            from_index = torch.full((x.shape[0],), from_index).to(x.device)

        if eps is None:
            eps = self.eps(diffused, from_index)

        return diffusion_space.decode(
            self.diffusion._predict_xstart_from_eps(x, from_index, eps)
        )

    def diffuse(self, images, to_index, noise=None):
        x0 = diffusion_space.encode(images)
        if isinstance(to_index, int) or to_index.ndim == 0:
            to_index = torch.full((x0.shape[0],), to_index).to(x0.device)
        if noise is None:
            noise = torch.randn_like(x0)
        assert noise.shape == x0.shape
        return diffusion_space.decode(self.diffusion.q_sample(x0, to_index, noise))

    def eps(self, diffused, from_index):
        x = diffusion_space.encode(diffused)
        if isinstance(from_index, int) or from_index.ndim == 0:
            from_index = torch.full((x.shape[0],), from_index).to(x.device)
        return self.model(x, from_index)[:, :3]

    def guided_eps(self, eps, grad, from_index, guidance_scale=1000):
        """
        Guided eps by differentiating through the diffusion model.

        Usage:

            diffused_image.requires_grad_(True)
            with torch.enable_grad():
                eps = diffusion.eps(diffused_image, from_index)
                denoised_image = diffusion.denoise(diffused_image, from_index, eps)

                augmentations_ = torch.cat([augmentations(denoised_image) for _ in range(4)])
                torch.stack(
                    [
                        (text_loss(augmentations_) / len(text_losses))
                        for text_loss in text_losses
                    ]
                ).mean().backward()

            eps = diffusion.guided_eps(eps, diffused_image.grad, from_index)
        """
        return (
            eps + guidance_scale * (1 - self.alphas_cumprod(from_index)).sqrt() * grad
        )

    def ts(self, index):
        return torch.tensor([index], device=self.device, dtype=torch.float32)

    def alphas_cumprod(self, index):
        return (
            torch.as_tensor(self.diffusion.alphas_cumprod[index])
            .float()
            .to(self.device)[None, None, None, None]
        )

    def sqrt_one_minus_alphas_cumprod(self, index):
        return (
            torch.as_tensor(self.diffusion.sqrt_one_minus_alphas_cumprod[index])
            .float()
            .to(self.device)[None, None, None, None]
        )

    def step(self, from_diffused, eps, from_index, to_index, noise=None, eta=1.0):
        if to_index > from_index:
            raise ValueError("to_index must be smaller than from_index")
        if noise is None:
            noise = torch.randn_like(eps)

        pred = diffusion_space.encode(self.denoise(from_diffused, from_index, eps))

        from_alphas_cumprod = self.alphas_cumprod(from_index)
        to_alphas_cumprod = self.alphas_cumprod(to_index)

        to_sigmas = eta * torch.sqrt(
            (1 - to_alphas_cumprod)
            / (1 - from_alphas_cumprod)
            * (1 - from_alphas_cumprod / to_alphas_cumprod)
        )

        dir_xt = (1.0 - to_alphas_cumprod - to_sigmas**2).sqrt() * eps
        to_x = to_alphas_cumprod.sqrt() * pred + dir_xt + to_sigmas * noise
        return diffusion_space.decode(to_x)


def create_openimages_model():
    model_config = model_and_diffusion_defaults()
    model_config.update(
        {
            "attention_resolutions": "32, 16, 8",
            "class_cond": False,
            # 'diffusion_steps': 1000, #No need to edit this, it is taken care of later.
            # 'rescale_timesteps': True,
            # 'timestep_respacing': 250, #No need to edit this, it is taken care of later.
            "image_size": 512,
            "learn_sigma": True,
            "noise_schedule": "linear",
            "num_channels": 256,
            "num_head_channels": 64,
            "num_res_blocks": 2,
            "resblock_updown": True,
            "use_checkpoint": True,
            "use_fp16": True,
            "use_scale_shift_norm": True,
        }
    )

    model, diffusion = create_model_and_diffusion(**model_config)
    if model_config["use_fp16"]:
        model.convert_to_fp16()
    return model, diffusion


def create_pixelart_model():
    model_config = model_and_diffusion_defaults()
    model_config.update(
        dict(
            image_size=256,
            learn_sigma=True,
            num_channels=128,
            num_res_blocks=2,
            num_heads=1,
            num_heads_upsample=-1,
            num_head_channels=-1,
            attention_resolutions="16",
            channel_mult="",
            dropout=0.0,
            class_cond=False,
            use_checkpoint=False,
            use_scale_shift_norm=False,
            resblock_updown=False,
            use_fp16=True,
            use_new_attention_order=False,
        )
    )

    model, diffusion = create_model_and_diffusion(**model_config)
    if model_config["use_fp16"]:
        model.convert_to_fp16()
    return model, diffusion


def create_model(
    image_size,
    num_channels,
    num_res_blocks,
    channel_mult="",
    learn_sigma=False,
    class_cond=False,
    use_checkpoint=False,
    attention_resolutions="16",
    num_heads=1,
    num_head_channels=-1,
    num_heads_upsample=-1,
    use_scale_shift_norm=False,
    dropout=0,
    resblock_updown=False,
    use_fp16=False,
    use_new_attention_order=False,
):
    if channel_mult == "":
        if image_size == 512:
            channel_mult = (0.5, 1, 1, 2, 2, 4, 4)
        elif image_size == 256:
            channel_mult = (1, 1, 2, 2, 4, 4)
        elif image_size == 128:
            channel_mult = (1, 1, 2, 3, 4)
        elif image_size == 64:
            channel_mult = (1, 2, 3, 4)
        else:
            raise ValueError(f"unsupported image size: {image_size}")
    else:
        channel_mult = tuple(int(ch_mult) for ch_mult in channel_mult.split(","))

    attention_ds = []
    for res in attention_resolutions.split(","):
        attention_ds.append(image_size // int(res))

    return UNetModel(
        image_size=image_size,
        in_channels=3,
        model_channels=num_channels,
        out_channels=(3 if not learn_sigma else 6),
        num_res_blocks=num_res_blocks,
        attention_resolutions=tuple(attention_ds),
        dropout=dropout,
        channel_mult=channel_mult,
        num_classes=None,
        use_checkpoint=use_checkpoint,
        use_fp16=use_fp16,
        num_heads=num_heads,
        num_head_channels=num_head_channels,
        num_heads_upsample=num_heads_upsample,
        use_scale_shift_norm=use_scale_shift_norm,
        resblock_updown=resblock_updown,
        use_new_attention_order=use_new_attention_order,
    )


def test_pixelart_diffusion():
    from perceptor import utils

    model = GuidedDiffusion("pixelart").cuda()
    diffused = model.random_diffused((1, 3, 256, 256))

    for from_index, to_index in model.schedule_indices(n_steps=50):
        eps = model.eps(diffused, from_index)
        diffused = model.step(diffused, eps, from_index, to_index)
    denoised = model.denoise(diffused, to_index)
    utils.pil_image(denoised).save("tests/pixelart.png")


def test_guided_diffusion_diffusion():
    from perceptor import utils

    model = GuidedDiffusion("standard").cuda()
    diffused = model.random_diffused((1, 3, 512, 512))

    for from_index, to_index in model.schedule_indices(n_steps=50):
        eps = model.eps(diffused, from_index)
        diffused = model.step(diffused, eps, from_index, to_index)
    denoised = model.denoise(diffused, to_index)
    utils.pil_image(denoised).save("tests/guided_diffusion.png")



"""
https://github.com/crowsonkb/deep-image-prior
https://github.com/DmitryUlyanov/deep-image-prior
"""
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from lantern import module_device, Tensor

from .get_hq_skip_net import get_hq_skip_net, get_offset_params, get_non_offset_params

DEFAULT_SIZE = 256
DEFAULT_SHAPE = (128, DEFAULT_SIZE, DEFAULT_SIZE)


class DeepImagePrior(nn.Module):
    def __init__(
        self,
        shape=DEFAULT_SHAPE,
        offset_type="none",
        n_scales=2,
        sigmoid=True,
        decorrelate_rgb=True,
        output_channels=3,
    ):
        """
        Represent image with latent and deep image prior network.

        Args:
            shape: (number of latent input channels, height, width)
            offset_type: type of offset to use, "1x1", "none", or "fulll"
            n_scales: number of up-downscales to use, if None, it is determined by the size of the input
            sigmoid: whether to use sigmoid activation
            decorrelate_rgb: whether to decorrelate the RGB channels
            output_channels: number of output channels
        """
        super().__init__()
        input_channels, height, width = shape
        assert height == width
        assert height % 8 == 0
        self.shape = shape
        self.n_scales = n_scales
        self.output_channels = output_channels
        self.network = get_hq_skip_net(
            input_depth=input_channels,
            offset_type=offset_type,
            num_scales=n_scales,
            decorr_rgb=decorrelate_rgb,
            sigmoid=sigmoid,
            n_channels=output_channels,
        )

    @property
    def device(self):
        return module_device(self)

    @property
    def input_channels(self):
        return self.shape[0]

    @property
    def height(self):
        return self.shape[1]

    @property
    def width(self):
        return self.shape[2]

    def forward(self, latents: Tensor.dims("NCHW")) -> Tensor.dims("NDHW"):
        return self.network(latents)

    def random_latents(self, size=1, n_channels=None):
        if n_channels is None:
            n_channels = self.input_channels
        return 0.1 * torch.randn(
            (size, n_channels, self.height, self.width), device=self.device
        )

    def fourier_latents(
        self,
        size=1,
        n_channels=None,
        min_log2_frequency=0.0,
        max_log2_frequency=9.0,
        log2_space=False,
    ):
        if n_channels is None:
            n_channels = self.input_channels
        assert n_channels % 4 == 0

        xs = torch.linspace(-1, 1, self.shape[2])
        ys = torch.linspace(-1, 1, self.shape[1])

        meshgrid = torch.stack(torch.meshgrid(xs, ys), dim=0)

        if log2_space:
            frequencies = 2.0 ** torch.linspace(
                min_log2_frequency, max_log2_frequency, steps=n_channels // 4
            )
        else:
            frequencies = torch.linspace(
                2.0**min_log2_frequency,
                2.0**max_log2_frequency,
                steps=n_channels // 4,
            )

        fourier_latents = torch.cat(
            [
                torch.sin(
                    meshgrid[None] * frequencies[:, None, None, None] * 2 * np.pi
                ),
                torch.cos(
                    meshgrid[None] * frequencies[:, None, None, None] * 2 * np.pi
                ),
            ],
            dim=0,
        ).flatten(end_dim=1)[None]
        return fourier_latents.to(self.device).repeat(size, 1, 1, 1).mul(0.3)

    def noisy_image_latents(self, images, n_channels=None, log_snr=-1.0):
        if n_channels is None:
            n_channels = self.input_channels

        sigma = 1 / (torch.as_tensor(log_snr).exp().sqrt() + 1)
        channels = images.shape[1]
        repeated_images = torch.stack(
            [images[:, index % channels] for index in range(n_channels)], dim=1
        )
        return 0.1 * (
            repeated_images.mul(2).sub(1) * (1 - sigma)
            + torch.randn_like(repeated_images) * sigma
        )

    def offset_parameters(self):
        return get_offset_params(self.network)

    def non_offset_parameters(self):
        return get_non_offset_params(self.network)

    def parameter_dicts(self, learning_rate):
        return [
            dict(
                params=self.offset_parameters(),
                lr=learning_rate * 0.1,
            ),
            dict(
                params=self.non_offset_parameters(),
                lr=learning_rate,
            ),
        ]


def test_deep_image_prior():
    model = DeepImagePrior().cuda()
    model(model.random_latents())



import torch

from . import utils


# DDPM/DDIM sampling


@torch.no_grad()
def sample(model, x, steps, eta=1.0, extra_args=dict()):
    """Draws samples from a model given starting noise."""
    ts = x.new_ones([x.shape[0]])

    # Create the noise schedule
    alphas, sigmas = utils.t_to_alpha_sigma(steps)

    # The sampling loop
    for i in range(len(steps)):

        # Get the model output (v, the predicted velocity)
        with torch.cuda.amp.autocast():
            v = model(x, ts * steps[i], **extra_args).float()

        # Predict the noise and the denoised image
        pred = x * alphas[i] - v * sigmas[i]
        eps = x * sigmas[i] + v * alphas[i]

        yield x, eps, pred

        # If we are not on the last timestep, compute the noisy image for the
        # next timestep.
        if i < len(steps) - 1:
            # If eta > 0, adjust the scaling factor for the predicted noise
            # downward according to the amount of additional noise to add
            ddim_sigma = (
                eta
                * (sigmas[i + 1] ** 2 / sigmas[i] ** 2).sqrt()
                * (1 - alphas[i] ** 2 / alphas[i + 1] ** 2).sqrt()
            )
            adjusted_sigma = (sigmas[i + 1] ** 2 - ddim_sigma**2).sqrt()

            # Recombine the predicted noise and predicted denoised image in the
            # correct proportions for the next step
            x = pred * alphas[i + 1] + eps * adjusted_sigma

            # Add the correct amount of fresh noise
            if eta:
                x += torch.randn_like(x) * ddim_sigma

    # If we are on the last timestep, output the denoised image
    # return pred


@torch.no_grad()
def reverse_sample(model, x, steps, extra_args, callback=None):
    """Finds a starting latent that would produce the given image with DDIM
    (eta=0) sampling."""
    ts = x.new_ones([x.shape[0]])

    # Create the noise schedule
    alphas, sigmas = utils.t_to_alpha_sigma(steps)

    # The sampling loop
    for i in range(len(steps) - 1):

        # Get the model output (v, the predicted velocity)
        with torch.cuda.amp.autocast():
            v = model(x, ts * steps[i], **extra_args).float()

        # Predict the noise and the denoised image
        pred = x * alphas[i] - v * sigmas[i]
        eps = x * sigmas[i] + v * alphas[i]

        # Call the callback
        if callback is not None:
            callback({"x": x, "i": i, "t": steps[i], "v": v, "pred": pred})

        # Recombine the predicted noise and predicted denoised image in the
        # correct proportions for the next step
        x = pred * alphas[i + 1] + eps * sigmas[i + 1]

    return x


# PNDM sampling (see https://openreview.net/pdf?id=PlKWVd2yBkY)


def make_eps_model_fn(model):
    def eps_model_fn(x, t, **extra_args):
        alphas, sigmas = utils.t_to_alpha_sigma(t)
        v = model(x, t, **extra_args)
        eps = x * sigmas[:, None, None, None] + v * alphas[:, None, None, None]
        return eps

    return eps_model_fn


def make_autocast_model_fn(model, enabled=True):
    def autocast_model_fn(*args, **kwargs):
        with torch.cuda.amp.autocast(enabled):
            return model(*args, **kwargs).float()

    return autocast_model_fn


def transfer(x, eps, t_1, t_2):
    alphas, sigmas = utils.t_to_alpha_sigma(t_1)
    next_alphas, next_sigmas = utils.t_to_alpha_sigma(t_2)
    pred = (x - eps * sigmas[:, None, None, None]) / alphas[:, None, None, None]
    x = pred * next_alphas[:, None, None, None] + eps * next_sigmas[:, None, None, None]
    return x, pred


def prk_step(model, x, t_1, t_2, extra_args):
    eps_model_fn = make_eps_model_fn(model)
    t_mid = (t_2 + t_1) / 2
    eps_1 = eps_model_fn(x, t_1, **extra_args)
    x_1, _ = transfer(x, eps_1, t_1, t_mid)
    eps_2 = eps_model_fn(x_1, t_mid, **extra_args)
    x_2, _ = transfer(x, eps_2, t_1, t_mid)
    eps_3 = eps_model_fn(x_2, t_mid, **extra_args)
    x_3, _ = transfer(x, eps_3, t_1, t_2)
    eps_4 = eps_model_fn(x_3, t_2, **extra_args)
    eps_prime = (eps_1 + 2 * eps_2 + 2 * eps_3 + eps_4) / 6
    x_new, pred = transfer(x, eps_prime, t_1, t_2)
    return x_new, eps_prime, pred


def plms_step(model, x, old_eps, t_1, t_2, extra_args):
    eps_model_fn = make_eps_model_fn(model)
    eps = eps_model_fn(x, t_1, **extra_args)
    eps_prime = (55 * eps - 59 * old_eps[-1] + 37 * old_eps[-2] - 9 * old_eps[-3]) / 24
    x_new, pred = transfer(x, eps_prime, t_1, t_2)
    return x_new, eps, pred


@torch.no_grad()
def prk_sample(model, x, steps, extra_args, is_reverse=False):
    """Draws samples from a model given starting noise using Pseudo Runge-Kutta."""
    ts = x.new_ones([x.shape[0]])
    model_fn = make_autocast_model_fn(model)
    if not is_reverse:
        steps = torch.cat([steps, steps.new_zeros([1])])
    for i in range(len(steps) - 1):
        x, eps, pred = prk_step(
            model_fn, x, steps[i] * ts, steps[i + 1] * ts, extra_args
        )
        yield x, eps, pred
    # return x if is_reverse else pred


@torch.no_grad()
def plms_sample(model, x, steps, extra_args, is_reverse=False):
    """Draws samples from a model given starting noise using Pseudo Linear Multistep."""
    ts = x.new_ones([x.shape[0]])
    model_fn = make_autocast_model_fn(model)
    if not is_reverse:
        steps = torch.cat([steps, steps.new_zeros([1])])
    old_eps = []
    for i in range(len(steps) - 1):
        if len(old_eps) < 3:
            x, eps, pred = prk_step(
                model_fn, x, steps[i] * ts, steps[i + 1] * ts, extra_args
            )
        else:
            x, eps, pred = plms_step(
                model_fn, x, old_eps, steps[i] * ts, steps[i + 1] * ts, extra_args
            )
            old_eps.pop(0)
        old_eps.append(eps)
        yield x, eps, pred
    # return x if is_reverse else pred
